import dask.array
import numpy
import orjson
import pytest
import xarray
import xarray.testing

from ..adapters.mapping import MapAdapter
from ..adapters.xarray import DatasetAdapter
from ..client import Context, from_context, record_history
from ..client import xarray as xarray_client
from ..serialization.xarray import serialize_json
from ..server.app import build_app

image = numpy.random.random((3, 5))
temp = 15 + 8 * numpy.random.randn(2, 2, 3)
precip = 10 * numpy.random.rand(2, 2, 3)
lon = [[-99.83, -99.32], [-99.79, -99.23]]
lat = [[42.25, 42.21], [42.63, 42.59]]

EXPECTED = {
    "image": xarray.Dataset(
        {
            "image": xarray.DataArray(
                xarray.Variable(
                    data=dask.array.from_array(image),
                    dims=["x", "y"],
                    attrs={"thing": "stuff"},
                ),
                coords={
                    "x": dask.array.arange(image.shape[0]) / 10,
                    "y": dask.array.arange(image.shape[1]) / 50,
                },
            ),
            "z": xarray.DataArray(data=dask.array.ones(17)),
        },
        attrs={"snow": "cold"},
    ),
    "weather": xarray.Dataset(
        {
            "temperature": (["x", "y", "time"], temp),
            "precipitation": (["x", "y", "time"], precip),
        },
        coords={
            "lon": (["x", "y"], lon),
            "lat": (["x", "y"], lat),
            "time": [1, 2, 3],
        },
    ),
    "wide": xarray.Dataset(
        {f"column_{i:03}": xarray.DataArray(i * numpy.ones(2)) for i in range(10)},
        coords={"time": numpy.arange(2)},
    ),
    "ragged": xarray.Dataset(
        {
            f"{i}": xarray.DataArray(i * numpy.ones(2 * i), dims=f"dim{i}")
            for i in range(3)
        }
    ),
}

tree = MapAdapter(
    {key: DatasetAdapter.from_dataset(ds) for key, ds in EXPECTED.items()}
)


@pytest.fixture(scope="module")
def client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


@pytest.mark.parametrize("key", list(tree))
def test_xarray_dataset(client, key):
    expected = EXPECTED[key]
    actual = client[key].read().load()
    xarray.testing.assert_equal(actual, expected)


@pytest.mark.parametrize("key", ["image", "weather"])
def test_dataset_column_access(client, key):
    expected_dataset = EXPECTED[key]
    actual_dataset = client[key].read().load()
    for col in expected_dataset:
        actual = actual_dataset[col]
        expected = expected_dataset[col]
        xarray.testing.assert_equal(actual, expected)


def test_wide_table_optimization(client):
    wide = client["wide"]
    with record_history() as history:
        wide.read()
    # This should be just a couple requests.
    # This upper bound is somewhat arbitrary to give wiggle room for future
    # minor changes. The point is: it's much less than one request per variable.
    assert len(history.requests) < 4


def test_wide_table_optimization_off(client):
    wide = client["wide"]
    with record_history() as history:
        wide.read(optimize_wide_table=False)
    assert len(history.requests) >= 10


def test_url_limit_handling(client):
    "Check that requests and split up to stay below the URL length limit."
    expected = EXPECTED["wide"]
    dsc = client["wide"]
    dsc.read()  # Dry run to run any one-off state-initializing requests.
    # Accumulate Requests here for later inspection.
    requests = []

    def accumulate(request):
        # httpx.AsyncClient requires event hooks to be async functions.
        requests.append(request)

    client.context.http_client.event_hooks["request"].append(accumulate)
    actual = dsc.read()
    xarray.testing.assert_equal(actual, expected)
    normal_request_count = len(requests)
    original = xarray_client.URL_CHARACTER_LIMIT
    try:
        # It should never be necessary to tune this for real-world use, but we
        # use this knob as a way to test its operation.
        xarray_client.URL_CHARACTER_LIMIT = 200
        # The client will need to split this across more requests in order to
        # stay within the tighter limit.
        requests.clear()  # Empty the Request cache before the next batch of requests.
        actual = dsc.read()
        xarray.testing.assert_equal(actual, expected)
        higher_request_count = len(requests)
        # Tighten even more.
        xarray_client.URL_CHARACTER_LIMIT = 100
        requests.clear()  # Empty the Request cache before the next batch of requests.
        actual = dsc.read()
        xarray.testing.assert_equal(actual, expected)
        highest_request_count = len(requests)
    finally:
        # Restore default.
        xarray_client.URL_CHARACTER_LIMIT = original
    # The goal here is to test the *trend* not the specific values because the
    # number of requests may evolve as the library changes, but the trend should
    # hold.
    assert highest_request_count > higher_request_count > normal_request_count


@pytest.mark.parametrize("ds_node", tree.values(), ids=tree.keys())
@pytest.mark.asyncio
async def test_serialize_json(ds_node: DatasetAdapter):
    """Verify that serialized Dataset keys are a subset
    of all coordinates and variables from the Dataset.
    Index variables are removed by serialize_json().
    """
    metadata = None  # Not used
    filter_for_access = None  # Not used
    result = await serialize_json(ds_node, metadata, filter_for_access)

    result_data_keys = orjson.loads(result).keys()
    ds_coords_and_vars = set(ds_node)

    assert set(result_data_keys).issubset(ds_coords_and_vars)



import base64
import io
from dataclasses import dataclass
from typing import List, Tuple, Union

B64_ENCODED_PREFIX = "data:application/vnd.apache.arrow.file;base64,"


@dataclass
class TableStructure:
    # This holds a Arrow schema, base64-encoded so that it can be transported
    # as JSON. For clarity, the encoded data (...) is prefixed like:
    #
    # data:application/vnd.apache.arrow.file;base64,...
    #
    # Arrow does not support an official JSON serialization, but it
    # could in the future: https://github.com/apache/arrow/pull/7110
    # If it does, we could switch to using that here.
    arrow_schema: str
    npartitions: int
    columns: List[str]
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_dask_dataframe(cls, ddf):
        import dask.dataframe.utils
        import pyarrow

        # Make a pandas Table with 0 rows.
        # We can use this to define an Arrow schema without loading any row data.
        meta = dask.dataframe.utils.make_meta(ddf)
        schema_bytes = pyarrow.Table.from_pandas(meta).schema.serialize()
        schema_b64 = base64.b64encode(schema_bytes).decode("utf-8")
        data_uri = B64_ENCODED_PREFIX + schema_b64
        return cls(
            arrow_schema=data_uri,
            npartitions=ddf.npartitions,
            columns=list(ddf.columns),
        )

    @classmethod
    def from_pandas(cls, df):
        import pyarrow

        schema_bytes = pyarrow.Table.from_pandas(df).schema.serialize()
        schema_b64 = base64.b64encode(schema_bytes).decode("utf-8")
        data_uri = B64_ENCODED_PREFIX + schema_b64
        return cls(arrow_schema=data_uri, npartitions=1, columns=list(df.columns))

    @property
    def arrow_schema_decoded(self):
        import pyarrow

        if not self.arrow_schema.startswith(B64_ENCODED_PREFIX):
            raise ValueError(
                f"Expected base64-encoded data prefixed with {B64_ENCODED_PREFIX}."
            )

        payload = self.arrow_schema[len(B64_ENCODED_PREFIX) :].encode(  # noqa: 203
            "utf-8"
        )
        return pyarrow.ipc.read_schema(io.BytesIO(base64.b64decode(payload)))

    @property
    def meta(self):
        return self.arrow_schema_decoded.empty_table().to_pandas()

    @classmethod
    def from_json(cls, content):
        return cls(
            arrow_schema=content["arrow_schema"],
            npartitions=content["npartitions"],
            columns=content["columns"],
            resizable=content["resizable"],
        )



import copy

from tiled.utils import Sentinel


def test_sentinel_self_and_copy_equality_and_identity():
    s = Sentinel("TEST")
    assert s == s
    assert s is s
    assert copy.copy(s) == s
    assert copy.copy(s) is s
    assert copy.deepcopy(s) == s
    assert copy.deepcopy(s) is s



import numpy as np
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..server.app import build_app
from ..structures.array import StructDtype


@pytest.mark.parametrize(
    "dtype,max_depth",
    [
        (np.dtype("u1, u2"), 1),
        (np.dtype([("a", "u1"), ("b", "f2")]), 1),
        (np.dtype([("a", [("b", "i"), ("c", "f")]), ("d", "c16", (2, 2))]), 2),
        (np.dtype([("a", [("b", "i", (1,)), ("c", "f")]), ("d", "c16", (2, 2))]), 2),
    ],
)
def test_dtype_rount_trip(dtype, max_depth):
    struct = StructDtype.from_numpy_dtype(dtype)
    assert dtype == struct.to_numpy_dtype()
    assert max_depth == struct.max_depth()


def test_fail_subtype():
    with pytest.raises(ValueError):
        StructDtype.from_numpy_dtype(np.dtype("8f"))


def test_read():
    data = np.array(
        [("Rex", 9, 81.0), ("Fido", 3, 27.0)],
        dtype=[("name", "U10"), ("age", "i4"), ("weight", "f4")],
    )
    tree = MapAdapter({"A": ArrayAdapter.from_array(data)})
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert np.all(data == client["A"].read())



from pathlib import Path

from ..catalog.register import DEFAULT_MIMETYPES_BY_FILE_EXT, resolve_mimetype


def test_resolve_mimetype():
    assert (
        resolve_mimetype(Path("test.csv"), DEFAULT_MIMETYPES_BY_FILE_EXT) == "text/csv"
    )



from pathlib import Path
from typing import Optional

import typer

from ._utils import get_context, get_profile  # noqa E402

admin_app = typer.Typer()


@admin_app.command("initialize-database")
def initialize_database(database_uri: str):
    """
    Initialize a SQL database for use by Tiled.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import UninitializedDatabase, check_database, stamp_head
    from ..authn_database.alembic_constants import (
        ALEMBIC_DIR,
        ALEMBIC_INI_TEMPLATE_PATH,
    )
    from ..authn_database.core import (
        ALL_REVISIONS,
        REQUIRED_REVISION,
        initialize_database,
    )

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        try:
            await check_database(engine, REQUIRED_REVISION, ALL_REVISIONS)
        except UninitializedDatabase:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} is new. Creating tables and marking revision {REQUIRED_REVISION}.",
                err=True,
            )
            await initialize_database(engine)
            typer.echo("Database initialized.", err=True)
        else:
            typer.echo(f"Database at {redacted_url} is already initialized.", err=True)
            raise typer.Abort()
        await engine.dispose()

    asyncio.run(do_setup())
    stamp_head(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri)


@admin_app.command("upgrade-database")
def upgrade_database(
    database_uri: str,
    revision: Optional[str] = typer.Argument(
        None,
        help="The ID of a revision to upgrade to. By default, upgrade to the latest one.",
    ),
):
    """
    Upgrade the database schema to the latest version.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import get_current_revision, upgrade
    from ..authn_database.alembic_constants import (
        ALEMBIC_DIR,
        ALEMBIC_INI_TEMPLATE_PATH,
    )
    from ..authn_database.core import ALL_REVISIONS

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        current_revision = await get_current_revision(engine, ALL_REVISIONS)
        await engine.dispose()
        if current_revision is None:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} has not been initialized. Use `tiled admin initialize-database`.",
                err=True,
            )
            raise typer.Abort()

    asyncio.run(do_setup())
    upgrade(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri, revision or "head")


@admin_app.command("downgrade-database")
def downgrade_database(
    database_uri: str,
    revision: str = typer.Argument(..., help="The ID of a revision to downgrade to."),
):
    """
    Upgrade the database schema to the latest version.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import downgrade, get_current_revision
    from ..authn_database.alembic_constants import (
        ALEMBIC_DIR,
        ALEMBIC_INI_TEMPLATE_PATH,
    )
    from ..authn_database.core import ALL_REVISIONS

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        current_revision = await get_current_revision(engine, ALL_REVISIONS)
        if current_revision is None:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} has not been initialized. Use `tiled admin initialize-database`.",
                err=True,
            )
            raise typer.Abort()

    asyncio.run(do_setup())
    downgrade(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri, revision)


@admin_app.command("check-config")
def check_config(
    config_path: Path = typer.Argument(
        None,
        help=(
            "Path to a config file or directory of config files. "
            "If None, check environment variable TILED_CONFIG. "
            "If that is unset, try default location ./config.yml."
        ),
    ),
):
    "Check configuration file for syntax and validation errors."
    import os

    from ..config import parse_configs

    config_path = config_path or os.getenv("TILED_CONFIG", "config.yml")
    try:
        parse_configs(config_path)
    except Exception as err:
        typer.echo(str(err), err=True)
        raise typer.Exit(1)
    typer.echo("No errors found in configuration.")


@admin_app.command("list-principals")
def list_principals(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    page_offset: int = typer.Argument(0),
    page_limit: int = typer.Argument(100, help="Max items to show"),
):
    """
    List information about all Principals (users or services) that have ever logged in.
    """
    import json

    context = get_context(profile)
    result = context.admin.list_principals(offset=page_offset, limit=page_limit)
    typer.echo(json.dumps(result, indent=2))


@admin_app.command("show-principal")
def show_principal(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    uuid: str = typer.Argument(..., help="UUID identifying Principal of interest"),
):
    """
    Show information about one Principal (user or service).
    """
    import json

    context = get_context(profile)
    result = context.admin.show_principal(uuid)
    typer.echo(json.dumps(result, indent=2))



from __future__ import annotations

import enum
import uuid
from datetime import datetime
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

import pydantic
import pydantic.dataclasses
import pydantic.errors
import pydantic.generics

from ..structures.core import StructureFamily
from .pydantic_array import ArrayStructure
from .pydantic_awkward import AwkwardStructure
from .pydantic_sparse import SparseStructure
from .pydantic_table import TableStructure

DataT = TypeVar("DataT")
LinksT = TypeVar("LinksT")
MetaT = TypeVar("MetaT")


class Error(pydantic.BaseModel):
    code: int
    message: str


class Response(pydantic.generics.GenericModel, Generic[DataT, LinksT, MetaT]):
    data: Optional[DataT]
    error: Optional[Error]
    links: Optional[LinksT]
    meta: Optional[MetaT]

    @pydantic.validator("error", always=True)
    def check_consistency(cls, v, values):
        if v is not None and values["data"] is not None:
            raise ValueError("must not provide both data and error")
        if v is None and values.get("data") is None:
            raise ValueError("must provide data or error")
        return v


class PaginationLinks(pydantic.BaseModel):
    self: str
    next: str
    prev: str
    first: str
    last: str


class EntryFields(str, enum.Enum):
    metadata = "metadata"
    structure_family = "structure_family"
    structure = "structure"
    count = "count"
    sorting = "sorting"
    specs = "specs"
    data_sources = "data_sources"
    none = ""


class NodeStructure(pydantic.BaseModel):
    contents: Optional[Dict[str, Resource[NodeAttributes, ResourceLinksT, EmptyDict]]]
    count: int


class SortingDirection(int, enum.Enum):
    ASCENDING = 1
    DECENDING = -1


class SortingItem(pydantic.BaseModel):
    key: str
    direction: SortingDirection


class Spec(pydantic.BaseModel, extra=pydantic.Extra.forbid, frozen=True):
    name: pydantic.constr(max_length=255)
    version: Optional[pydantic.constr(max_length=255)]


# Wait for fix https://github.com/pydantic/pydantic/issues/3957
# Specs = pydantic.conlist(Spec, max_items=20, unique_items=True)
Specs = pydantic.conlist(Spec, max_items=20)


class Asset(pydantic.BaseModel):
    data_uri: str
    is_directory: bool
    parameter: Optional[str]
    num: Optional[int] = None
    id: Optional[int] = None

    @classmethod
    def from_orm(cls, orm):
        return cls(
            data_uri=orm.asset.data_uri,
            is_directory=orm.asset.is_directory,
            parameter=orm.parameter,
            num=orm.num,
            id=orm.asset.id,
        )


class Management(str, enum.Enum):
    external = "external"
    immutable = "immutable"
    locked = "locked"
    writable = "writable"


class Revision(pydantic.BaseModel):
    revision_number: int
    metadata: dict
    specs: Specs
    time_updated: datetime

    @classmethod
    def from_orm(cls, orm):
        # Trailing underscore in 'metadata_' avoids collision with
        # SQLAlchemy reserved word 'metadata'.
        return cls(
            revision_number=orm.revision_number,
            metadata=orm.metadata_,
            specs=orm.specs,
            time_updated=orm.time_updated,
        )


class DataSource(pydantic.BaseModel):
    id: Optional[int] = None
    structure: Optional[
        Union[
            ArrayStructure,
            AwkwardStructure,
            TableStructure,
            NodeStructure,
            SparseStructure,
        ]
    ] = None
    mimetype: Optional[str] = None
    parameters: dict = {}
    assets: List[Asset] = []
    management: Management = Management.writable

    @classmethod
    def from_orm(cls, orm):
        return cls(
            id=orm.id,
            structure=getattr(orm.structure, "structure", None),
            mimetype=orm.mimetype,
            parameters=orm.parameters,
            assets=[Asset.from_orm(assoc) for assoc in orm.asset_associations],
            management=orm.management,
        )


class NodeAttributes(pydantic.BaseModel):
    ancestors: List[str]
    structure_family: Optional[StructureFamily]
    specs: Optional[Specs]
    metadata: Optional[Dict]  # free-form, user-specified dict
    structure: Optional[
        Union[
            ArrayStructure,
            AwkwardStructure,
            TableStructure,
            NodeStructure,
            SparseStructure,
        ]
    ]
    sorting: Optional[List[SortingItem]]
    data_sources: Optional[List[DataSource]]


AttributesT = TypeVar("AttributesT")
ResourceMetaT = TypeVar("ResourceMetaT")
ResourceLinksT = TypeVar("ResourceLinksT")


class SelfLinkOnly(pydantic.BaseModel):
    self: str


class ContainerLinks(pydantic.BaseModel):
    self: str
    search: str
    full: str


class ArrayLinks(pydantic.BaseModel):
    self: str
    full: str
    block: str


class AwkwardLinks(pydantic.BaseModel):
    self: str
    buffers: str
    full: str


class DataFrameLinks(pydantic.BaseModel):
    self: str
    full: str
    partition: str


class SparseLinks(pydantic.BaseModel):
    self: str
    full: str
    block: str


resource_links_type_by_structure_family = {
    StructureFamily.container: ContainerLinks,
    StructureFamily.array: ArrayLinks,
    StructureFamily.awkward: AwkwardLinks,
    StructureFamily.table: DataFrameLinks,
    StructureFamily.sparse: SparseLinks,
}


class EmptyDict(pydantic.BaseModel):
    pass


class ContainerMeta(pydantic.BaseModel):
    count: int


class Resource(
    pydantic.generics.GenericModel, Generic[AttributesT, ResourceLinksT, ResourceMetaT]
):
    "A JSON API Resource"
    id: Union[str, uuid.UUID]
    attributes: AttributesT
    links: Optional[ResourceLinksT]
    meta: Optional[ResourceMetaT]


class AccessAndRefreshTokens(pydantic.BaseModel):
    access_token: str
    expires_in: int
    refresh_token: str
    refresh_token_expires_in: int
    token_type: str


class RefreshToken(pydantic.BaseModel):
    refresh_token: str


class DeviceCode(pydantic.BaseModel):
    device_code: str
    grant_type: str


class AuthenticationMode(str, enum.Enum):
    password = "password"
    external = "external"


class AboutAuthenticationProvider(pydantic.BaseModel):
    provider: str
    mode: AuthenticationMode
    links: Dict[str, str]
    confirmation_message: Optional[str]


class AboutAuthenticationLinks(pydantic.BaseModel):
    whoami: str
    apikey: str
    refresh_session: str
    revoke_session: str
    logout: str


class AboutAuthentication(pydantic.BaseModel):
    required: bool
    providers: List[AboutAuthenticationProvider]
    links: Optional[AboutAuthenticationLinks]


class About(pydantic.BaseModel):
    api_version: int
    library_version: str
    formats: Dict[str, List[str]]
    aliases: Dict[str, Dict[str, List[str]]]
    queries: List[str]
    authentication: AboutAuthentication
    links: Dict[str, str]
    meta: Dict


class PrincipalType(str, enum.Enum):
    user = "user"
    service = "service"


class Identity(pydantic.BaseModel, orm_mode=True):
    id: pydantic.constr(max_length=255)
    provider: pydantic.constr(max_length=255)
    latest_login: Optional[datetime]


class Role(pydantic.BaseModel, orm_mode=True):
    name: str
    scopes: List[str]
    # principals


class APIKey(pydantic.BaseModel, orm_mode=True):
    first_eight: pydantic.constr(min_length=8, max_length=8)
    expiration_time: Optional[datetime]
    note: Optional[pydantic.constr(max_length=255)]
    scopes: List[str]
    latest_activity: Optional[datetime] = None


class APIKeyWithSecret(APIKey):
    secret: str  # hex-encoded bytes

    @classmethod
    def from_orm(cls, orm, secret):
        return cls(
            first_eight=orm.first_eight,
            expiration_time=orm.expiration_time,
            note=orm.note,
            scopes=orm.scopes,
            latest_activity=orm.latest_activity,
            secret=secret,
        )


class Session(pydantic.BaseModel, orm_mode=True):
    """
    This related to refresh tokens, which have a session uuid ("sid") claim.

    When the client attempts to use a refresh token, we first check
    here to ensure that the "session", which is associated with a chain
    of refresh tokens that came from a single authentication, are still valid.
    """

    # The id field (primary key) is intentionally not exposed to the application.
    # It is left as an internal database concern.
    uuid: uuid.UUID
    expiration_time: datetime
    revoked: bool


class Principal(pydantic.BaseModel, orm_mode=True):
    "Represents a User or Service"
    # The id field (primary key) is intentionally not exposed to the application.
    # It is left as an internal database concern.
    uuid: uuid.UUID
    type: PrincipalType
    identities: List[Identity] = []
    roles: List[Role] = []
    api_keys: List[APIKey] = []
    sessions: List[Session] = []
    latest_activity: Optional[datetime] = None

    @classmethod
    def from_orm(cls, orm, latest_activity=None):
        instance = super().from_orm(orm)
        instance.latest_activity = latest_activity
        return instance


class APIKeyRequestParams(pydantic.BaseModel):
    # Provide an example for expires_in. Otherwise, OpenAPI suggests lifetime=0.
    # If the user is not reading carefully, they will be frustrated when they
    # try to use the instantly-expiring API key!
    expires_in: Optional[int] = pydantic.Field(..., example=600)  # seconds
    scopes: Optional[List[str]] = pydantic.Field(..., example=["inherit"])
    note: Optional[str]


class PostMetadataRequest(pydantic.BaseModel):
    id: Optional[str] = None
    structure_family: StructureFamily
    metadata: Dict = {}
    data_sources: List[DataSource] = []
    specs: Specs = []

    # Wait for fix https://github.com/pydantic/pydantic/issues/3957
    # to do this with `unique_items` parameters to `pydantic.constr`.
    @pydantic.validator("specs", always=True)
    def specs_uniqueness_validator(cls, v):
        if v is None:
            return None
        for i, value in enumerate(v, start=1):
            if value in v[i:]:
                raise pydantic.errors.ListUniqueItemsError()
        return v


class PostMetadataResponse(pydantic.BaseModel, Generic[ResourceLinksT]):
    id: str
    links: Union[ArrayLinks, DataFrameLinks, SparseLinks]
    metadata: Dict
    data_sources: List[DataSource]


class PutMetadataResponse(pydantic.BaseModel, Generic[ResourceLinksT]):
    id: str
    links: Union[ArrayLinks, DataFrameLinks, SparseLinks]
    # May be None if not altered
    metadata: Optional[Dict]
    data_sources: Optional[List[DataSource]]


class DistinctValueInfo(pydantic.BaseModel):
    value: Any
    count: Optional[int]


class GetDistinctResponse(pydantic.BaseModel):
    metadata: Optional[Dict[str, List[DistinctValueInfo]]]
    structure_families: Optional[List[DistinctValueInfo]]
    specs: Optional[List[DistinctValueInfo]]


class PutMetadataRequest(pydantic.BaseModel):
    # These fields are optional because None means "no changes; do not update".
    metadata: Optional[Dict]
    specs: Optional[Specs]

    # Wait for fix https://github.com/pydantic/pydantic/issues/3957
    # to do this with `unique_items` parameters to `pydantic.constr`.
    @pydantic.validator("specs", always=True)
    def specs_uniqueness_validator(cls, v):
        if v is None:
            return None
        for i, value in enumerate(v, start=1):
            if value in v[i:]:
                raise pydantic.errors.ListUniqueItemsError()
        return v


NodeStructure.update_forward_refs()



"""Separate structure table

Revision ID: 2ca16566d692
Revises: 1cd99c02d0c7
Create Date: 2024-01-22 20:44:23.132801

"""
from datetime import datetime

import orjson
import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import func

from tiled.catalog.orm import JSONVariant
from tiled.catalog.utils import compute_structure_id
from tiled.server.schemas import Management

# revision identifiers, used by Alembic.
revision = "2ca16566d692"
down_revision = "1cd99c02d0c7"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()

    # Create new 'structures' table.
    op.create_table(
        "structures",
        sa.Column("id", sa.Unicode(32), primary_key=True, unique=True),
        sa.Column("structure", JSONVariant, nullable=False),
    )
    # Get reference, to be used for copying data.
    structures = sa.Table(
        "structures",
        sa.MetaData(),
        sa.Column("id", sa.Unicode(32)),
        sa.Column("structure", JSONVariant),
    )
    if connection.engine.dialect.name == "sqlite":
        # We use a copy-and-move strategy here because we cannot get exactly
        # the result we want by adding a FOREIGN KEY to SQLite on an existing
        # table.
        op.create_table(
            "new_data_sources",
            sa.Column(
                "id", sa.Integer, primary_key=True, index=True, autoincrement=True
            ),
            sa.Column(
                "node_id",
                sa.Integer,
                sa.ForeignKey("nodes.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column(
                "structure_id",
                sa.Unicode(32),
                sa.ForeignKey("structures.id", ondelete="CASCADE"),
                nullable=True,
            ),
            sa.Column("mimetype", sa.Unicode(255), nullable=False),
            sa.Column("parameters", JSONVariant, nullable=True),
            sa.Column("management", sa.Enum(Management), nullable=False),
            sa.Column(
                "time_created",
                sa.DateTime(timezone=False),
                server_default=func.now(),
            ),
            sa.Column(
                "time_updated",
                sa.DateTime(timezone=False),
                onupdate=func.now(),
                server_default=func.now(),
            ),
        )
        new_data_sources = sa.Table(
            "new_data_sources",
            sa.MetaData(),
            sa.Column("id", sa.Integer),
            sa.Column(
                "node_id",
                sa.Integer,
                sa.ForeignKey("nodes.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column("structure_id", sa.Unicode(32)),
            sa.Column("mimetype", sa.Unicode(255), nullable=False),
            sa.Column("parameters", JSONVariant, nullable=True),
            sa.Column("management", sa.Enum(Management), nullable=False),
            sa.Column(
                "time_created",
                sa.DateTime(timezone=False),
                server_default=func.now(),
            ),
            sa.Column(
                "time_updated",
                sa.DateTime(timezone=False),
                onupdate=func.now(),
                server_default=func.now(),
            ),
        )
        # Extract rows from data_sources and compute structure_id.
        results = connection.execute(
            sa.text(
                "SELECT id, node_id, structure, mimetype, parameters, management, "
                "time_created, time_updated FROM data_sources"
            )
        ).fetchall()

        new_data_sources_rows = []
        unique_structures = {}  # map unique structure_id -> structure
        for row in results:
            structure_id = compute_structure_id(row[2])
            new_row = {
                "id": row[0],
                "node_id": row[1],
                "structure_id": structure_id,
                "mimetype": row[3],
                "parameters": orjson.loads(row[4]),
                "management": row[5],
                "time_created": datetime.fromisoformat(row[6]),
                "time_udpated": datetime.fromisoformat(row[7]),
            }
            new_data_sources_rows.append(new_row)
            if structure_id not in unique_structures:
                unique_structures[structure_id] = orjson.loads(row[2])
        structures_rows = [
            {"id": structure_id, "structure": structure}
            for structure_id, structure in unique_structures.items()
        ]

        # Copy data into new tables.
        op.bulk_insert(structures, structures_rows)
        op.bulk_insert(new_data_sources, new_data_sources_rows)

        # Drop old 'data_structures' and move 'new_data_structures' into its place.
        op.drop_table("data_sources")
        op.rename_table("new_data_sources", "data_sources")
        # The above leaves many partially filled pages and, run on example
        # datasets, left the database slightly _larger_. Clean up.
        with op.get_context().autocommit_block():
            connection.execute(sa.text("VACUUM"))
    else:
        # PostgreSQL
        # Extract rows from data_sources and compute structure_id.
        results = connection.execute(
            sa.text("SELECT id, structure FROM data_sources")
        ).fetchall()
        unique_structures = {}  # map unique structure_id -> structure
        data_source_id_to_structure_id = {}
        for data_source_id, structure in results:
            structure_id = compute_structure_id(structure)
            unique_structures[structure_id] = structure
            data_source_id_to_structure_id[data_source_id] = structure_id
        structures_rows = [
            {"id": structure_id, "structure": structure}
            for structure_id, structure in unique_structures.items()
        ]
        # Copy data into 'structures' table.
        op.bulk_insert(structures, structures_rows)
        op.add_column(
            "data_sources",
            sa.Column(
                "structure_id",
                sa.Unicode(32),
                sa.ForeignKey("structures.id", ondelete="CASCADE"),
            ),
        )
        data_sources = sa.Table(
            "data_sources",
            sa.MetaData(),
            sa.Column("id", sa.Integer),
            sa.Column(
                "node_id",
                sa.Integer,
                sa.ForeignKey("nodes.id", ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column("structure_id", sa.Unicode(32)),
            sa.Column("structure", sa.Unicode(32)),
            sa.Column("mimetype", sa.Unicode(255), nullable=False),
            sa.Column("parameters", JSONVariant, nullable=True),
            sa.Column("management", sa.Enum(Management), nullable=False),
        )
        for data_source_id, structure_id in data_source_id_to_structure_id.items():
            connection.execute(
                data_sources.update()
                .values(structure_id=structure_id)
                .where(data_sources.c.id == data_source_id)
            )
        op.drop_column("data_sources", "structure")


def downgrade():
    # This _could_ be implemented but we will wait for a need since we are
    # still in alpha releases.
    raise NotImplementedError



import itertools

import dask
import dask.array
import numpy

from .base import BaseClient
from .utils import export_util, handle_error, params_from_slice


class _DaskArrayClient(BaseClient):
    "Client-side wrapper around an array-like that returns dask arrays"

    def __init__(self, *args, item, **kwargs):
        super().__init__(*args, item=item, **kwargs)

    @property
    def dims(self):
        return self.structure().dims

    @property
    def shape(self):
        return self.structure().shape

    @property
    def size(self):
        return numpy.product(self.structure().shape)

    @property
    def dtype(self):
        return self.structure().data_type.to_numpy_dtype()

    @property
    def nbytes(self):
        structure = self.structure()
        return (
            numpy.product(structure.shape)
            * structure.data_type.to_numpy_dtype().itemsize
        )

    @property
    def chunks(self):
        return self.structure().chunks

    @property
    def ndim(self):
        return len(self.structure().shape)

    def __repr__(self):
        structure = self.structure()
        attrs = {
            "shape": structure.shape,
            "chunks": structure.chunks,
            "dtype": structure.data_type.to_numpy_dtype(),
        }
        if structure.dims:
            attrs["dims"] = structure.dims
        return (
            f"<{type(self).__name__}"
            + "".join(f" {k}={v}" for k, v in attrs.items())
            + ">"
        )

    def __array__(self, *args, **kwargs):
        return self.read().__array__(*args, **kwargs)

    def _get_block(self, block, dtype, shape, slice=None):
        """
        Fetch the actual data for one block in a chunked (dask) array.

        See read_block() for a public version of this. This private version
        enables more efficient multi-block access by requiring the caller to
        pass in the structure (dtype, shape).
        """
        media_type = "application/octet-stream"
        if slice is not None:
            # TODO The server accepts a slice parameter but we'll need to write
            # careful code here to determine what the new shape will be.
            raise NotImplementedError(
                "Slicing less than one block is not yet supported."
            )
        if shape:
            # Check for special case of shape with 0 in it.
            if 0 in shape:
                # This is valid, and it has come up in the wild.  An array with
                # 0 as one of the dimensions never contains data, so we can
                # short-circuit here without any further information from the
                # service.
                return numpy.array([], dtype=dtype).reshape(shape)
            expected_shape = ",".join(map(str, shape))
        else:
            expected_shape = "scalar"
        content = handle_error(
            self.context.http_client.get(
                self.item["links"]["block"],
                headers={"Accept": media_type},
                params={
                    "block": ",".join(map(str, block)),
                    "expected_shape": expected_shape,
                },
            )
        ).read()
        return numpy.frombuffer(content, dtype=dtype).reshape(shape)

    def read_block(self, block, slice=None):
        """
        Access the data for one block of this chunked (dask) array.

        Optionally, access only a slice *within* this block.
        """
        structure = self.structure()
        chunks = structure.chunks
        dtype = structure.data_type.to_numpy_dtype()
        try:
            shape = tuple(chunks[dim][i] for dim, i in enumerate(block))
        except IndexError:
            raise IndexError(f"Block index {block} out of range")
        dask_array = dask.array.from_delayed(
            dask.delayed(self._get_block)(block, dtype, shape), dtype=dtype, shape=shape
        )
        # TODO Make the request in _get_block include the slice so that we only
        # fetch exactly the data that we want. This will require careful code
        # to determine what the shape will be.
        if slice is not None:
            dask_array = dask_array[slice]
        return dask_array

    def read(self, slice=None):
        """
        Acess the entire array or a slice.

        The array will be internally chunked with dask.
        """
        structure = self.structure()
        shape = structure.shape
        dtype = structure.data_type.to_numpy_dtype()
        # Build a client-side dask array whose chunks pull from a server-side
        # dask array.
        name = "remote-dask-array-" f"{self.uri}"
        chunks = structure.chunks
        # Count the number of blocks along each axis.
        num_blocks = (range(len(n)) for n in chunks)
        # Loop over each block index --- e.g. (0, 0), (0, 1), (0, 2) .... ---
        # and build a dask task encoding the method for fetching its data from
        # the server.
        dask_tasks = {
            (name,)
            + block: (
                self._get_block,
                block,
                dtype,
                tuple(chunks[dim][i] for dim, i in enumerate(block)),
            )
            for block in itertools.product(*num_blocks)
        }
        dask_array = dask.array.Array(
            dask=dask_tasks, name=name, chunks=chunks, dtype=dtype, shape=shape
        )
        if slice is not None:
            dask_array = dask_array[slice]
        return dask_array

    def write(self, array):
        handle_error(
            self.context.http_client.put(
                self.item["links"]["full"],
                content=array.tobytes(),
                headers={"Content-Type": "application/octet-stream"},
            )
        )

    def write_block(self, array, block):
        handle_error(
            self.context.http_client.put(
                self.item["links"]["block"].format(*block),
                content=array.tobytes(),
                headers={"Content-Type": "application/octet-stream"},
            )
        )

    def __getitem__(self, slice):
        return self.read(slice)

    # The default object.__iter__ works as expected here, no need to
    # implemented it specifically.

    def __len__(self):
        # As with numpy, len(arr) is the size of the zeroth axis.
        return self.structure().shape[0]

    def export(
        self, filepath, *, format=None, slice=None, link=None, template_vars=None
    ):
        """
        Download data in some format and write to a file.

        Parameters
        ----------
        file: str or buffer
            Filepath or writeable buffer.
        format : str, optional
            If format is None and `file` is a filepath, the format is inferred
            from the name, like 'table.csv' implies format="text/csv". The format
            may be given as a file extension ("csv") or a media type ("text/csv").
        slice : List[slice], optional
            List of slice objects. A convenient way to generate these is shown
            in the examples.
        link: str, optional
            Used internally. Refers to a key in the dictionary of links sent
            from the server.
        template_vars: dict, optional
            Used internally.

        Examples
        --------

        Export all.

        >>> a.export("numbers.csv")

        Export an N-dimensional slice.

        >>> import numpy
        >>> a.export("numbers.csv", slice=numpy.s_[:10, 50:100])
        """
        # For array, this is always 'full', but xarray clients set a custom link.
        if link is None:
            link = "full"
        template_vars = template_vars or {}
        params = params_from_slice(slice)
        return export_util(
            filepath,
            format,
            self.context.http_client.get,
            self.item["links"][link].format(**template_vars),
            params=params,
        )


# Subclass with a public class that adds the dask-specific methods.


class DaskArrayClient(_DaskArrayClient):
    "Client-side wrapper around an array-like that returns dask arrays"

    def compute(self):
        "Alias to client.read().compute()"
        return self.read().compute()


class ArrayClient(DaskArrayClient):
    "Client-side wrapper around an array-like that returns in-memory arrays"

    def read(self, slice=None):
        """
        Acess the entire array or a slice.
        """
        return super().read(slice).compute()

    def read_block(self, block, slice=None):
        """
        Access the data for one block of this chunked array.

        Optionally, access only a slice *within* this block.
        """
        return super().read_block(block, slice).compute()



import collections
import importlib
import operator
import os
import re
import shutil
import sys
import uuid
from functools import partial, reduce
from pathlib import Path
from urllib.parse import quote_plus, urlparse

import anyio
from fastapi import HTTPException
from sqlalchemy import delete, event, func, not_, or_, select, text, type_coerce, update
from sqlalchemy.dialects.postgresql import JSONB, REGCONFIG
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import selectinload
from sqlalchemy.sql.expression import cast

from tiled.queries import (
    Comparison,
    Contains,
    Eq,
    FullText,
    In,
    KeysFilter,
    NotEq,
    NotIn,
    Operator,
    SpecsQuery,
    StructureFamilyQuery,
)

from ..query_registration import QueryTranslationRegistry
from ..server.schemas import DataSource, Management, Revision, Spec
from ..structures.core import StructureFamily
from ..utils import (
    UNCHANGED,
    Conflicts,
    OneShotCachedMap,
    UnsupportedQueryType,
    ensure_awaitable,
    import_object,
    path_from_uri,
    safe_json_dump,
)
from . import orm
from .core import check_catalog_database, initialize_database
from .explain import ExplainAsyncSession
from .mimetypes import (
    AWKWARD_BUFFERS_MIMETYPE,
    DEFAULT_ADAPTERS_BY_MIMETYPE,
    PARQUET_MIMETYPE,
    SPARSE_BLOCKS_PARQUET_MIMETYPE,
    ZARR_MIMETYPE,
)
from .utils import SCHEME_PATTERN, compute_structure_id, ensure_uri

DEFAULT_ECHO = bool(int(os.getenv("TILED_ECHO_SQL", "0") or "0"))
INDEX_PATTERN = re.compile(r"^[A-Za-z0-9_-]+$")

# When data is uploaded, how is it saved?
# TODO: Make this configurable at Catalog construction time.
DEFAULT_CREATION_MIMETYPE = {
    StructureFamily.array: ZARR_MIMETYPE,
    StructureFamily.awkward: AWKWARD_BUFFERS_MIMETYPE,
    StructureFamily.table: PARQUET_MIMETYPE,
    StructureFamily.sparse: SPARSE_BLOCKS_PARQUET_MIMETYPE,
}
DEFAULT_INIT_STORAGE = OneShotCachedMap(
    {
        StructureFamily.array: lambda: importlib.import_module(
            "...adapters.zarr", __name__
        ).ZarrArrayAdapter.init_storage,
        StructureFamily.awkward: lambda: importlib.import_module(
            "...adapters.awkward_buffers", __name__
        ).AwkwardBuffersAdapter.init_storage,
        StructureFamily.table: lambda: importlib.import_module(
            "...adapters.parquet", __name__
        ).ParquetDatasetAdapter.init_storage,
        StructureFamily.sparse: lambda: importlib.import_module(
            "...adapters.sparse_blocks_parquet", __name__
        ).SparseBlocksParquetAdapter.init_storage,
    }
)


class RootNode:
    """
    Node representing the root of the tree.

    This node is special because the state configuring this node arises from
    server initialization (typically a configuration file) not from the
    database.

    It mocks the relevant part of the interface of .orm.Node.
    """

    structure_family = StructureFamily.container

    def __init__(self, metadata, specs, access_policy):
        self.metadata_ = metadata or {}
        self.specs = [Spec.parse_obj(spec) for spec in specs or []]
        self.ancestors = []
        self.key = None
        self.data_sources = None


class Context:
    def __init__(
        self,
        engine,
        writable_storage=None,
        readable_storage=None,
        adapters_by_mimetype=None,
        key_maker=lambda: str(uuid.uuid4()),
    ):
        self.engine = engine
        readable_storage = readable_storage or []
        if not isinstance(readable_storage, list):
            raise ValueError("readable_storage should be a list of URIs or paths")
        if writable_storage:
            writable_storage = ensure_uri(str(writable_storage))
            if not urlparse(writable_storage).scheme == "file":
                raise NotImplementedError(
                    "Only file://... writable storage is currently supported."
                )
            # If it is writable, it is automatically also readable.
            readable_storage.append(writable_storage)
        self.writable_storage = writable_storage
        self.readable_storage = [ensure_uri(path) for path in readable_storage]
        self.key_maker = key_maker
        adapters_by_mimetype = adapters_by_mimetype or {}
        # If adapters_by_mimetype comes from a configuration file,
        # objects are given as importable strings, like "package.module:Reader".
        for key, value in list(adapters_by_mimetype.items()):
            if isinstance(value, str):
                adapters_by_mimetype[key] = import_object(value)
        # User-provided adapters take precedence over defaults.
        merged_adapters_by_mimetype = collections.ChainMap(
            adapters_by_mimetype, DEFAULT_ADAPTERS_BY_MIMETYPE
        )
        self.adapters_by_mimetype = merged_adapters_by_mimetype

    def session(self):
        "Convenience method for constructing an AsyncSession context"
        return ExplainAsyncSession(self.engine, autoflush=False, expire_on_commit=False)

    # This is heading in a reasonable direction but does not actually work yet.
    # Pausing development for now.
    #
    #     async def list_metadata_indexes(self):
    #         dialect_name = self.engine.url.get_dialect().name
    #         async with self.context.session() as db:
    #             if dialect_name == "sqlite":
    #                 statement = """
    # SELECT name, sql
    # FROM SQLite_master
    # WHERE type = 'index'
    # AND tbl_name = 'nodes'
    # AND name LIKE 'tiled_md_%'
    # ORDER BY tbl_name;
    #
    # """
    #             elif dialect_name == "postgresql":
    #                 statement = """
    # SELECT
    # indexname, indexdef
    # FROM
    # pg_indexes
    # WHERE tablename = 'nodes'
    # AND indexname LIKE 'tiled_md_%'
    # ORDER BY
    # indexname;
    #
    # """
    #             else:
    #                 raise NotImplementedError(
    #                     f"Cannot list indexes for dialect {dialect_name}"
    #                 )
    #             indexes = (
    #                 await db.execute(
    #                     text(statement),
    #                     explain=False,
    #                 )
    #             ).all()
    #         return indexes
    #
    #     async def create_metadata_index(self, index_name, key):
    #         """
    #
    #         References
    #         ----------
    #         https://scalegrid.io/blog/using-jsonb-in-postgresql-how-to-effectively-store-index-json-data-in-postgresql/
    #         https://pganalyze.com/blog/gin-index
    #         """
    #         dialect_name = self.engine.url.get_dialect().name
    #         if INDEX_PATTERN.match(index_name) is None:
    #             raise ValueError(f"Index name must match pattern {INDEX_PATTERN}")
    #         if dialect_name == "sqlite":
    #             expression = orm.Node.metadata_[key].as_string()
    #         elif dialect_name == "postgresql":
    #             expression = orm.Node.metadata_[key].label("md")
    #         else:
    #             raise NotImplementedError
    #         index = Index(
    #             f"tiled_md_{index_name}",
    #             "ancestors",
    #             expression,
    #             # postgresql_ops={"md": "jsonb_ops"},
    #             postgresql_using="gin",
    #         )
    #
    #         def create_index(connection):
    #             index.create(connection)
    #
    #         async with self.engine.connect() as connection:
    #             await connection.run_sync(create_index)
    #             await connection.commit()
    #
    #     async def drop_metadata_index(self, index_name):
    #         if INDEX_PATTERN.match(index_name) is None:
    #             raise ValueError(f"Index name must match pattern {INDEX_PATTERN}")
    #         if not index_name.startswith("tiled_md_"):
    #             index_name = f"tiled_md_{index_name}"
    #         async with self.context.session() as db:
    #             await db.execute(text(f"DROP INDEX {index_name};"), explain=False)
    #             await db.commit()
    #
    #     async def drop_all_metadata_indexes(self):
    #         indexes = await self.list_metadata_indexes()
    #         async with self.context.session() as db:
    #             for index_name, sql in indexes:
    #                 await db.execute(text(f"DROP INDEX {index_name};"), explain=False)
    #             await db.commit()

    async def execute(self, statement, explain=None):
        "Debugging convenience utility, not exposed to server"
        async with self.session() as db:
            result = await db.execute(text(statement), explain=explain)
            await db.commit()
            return result


class CatalogNodeAdapter:
    query_registry = QueryTranslationRegistry()
    register_query = query_registry.register
    register_query_lazy = query_registry.register_lazy

    def __init__(
        self,
        context,
        node,
        *,
        conditions=None,
        queries=None,
        sorting=None,
        access_policy=None,
    ):
        self.context = context
        self.engine = self.context.engine
        self.node = node
        if node.key is None:
            # Special case for RootNode
            self.segments = []
        else:
            self.segments = node.ancestors + [node.key]
        self.sorting = sorting or [("", 1)]
        self.order_by_clauses = order_by_clauses(self.sorting)
        self.conditions = conditions or []
        self.queries = queries or []
        self.structure_family = node.structure_family
        self.specs = [Spec.parse_obj(spec) for spec in node.specs]
        self.ancestors = node.ancestors
        self.key = node.key
        self.access_policy = access_policy
        self.startup_tasks = [self.startup]
        self.shutdown_tasks = [self.shutdown]

    def metadata(self):
        return self.node.metadata_

    async def startup(self):
        if (self.context.engine.dialect.name == "sqlite") and (
            self.context.engine.url.database == ":memory:"
        ):
            # Special-case for in-memory SQLite: Because it is transient we can
            # skip over anything related to migrations.
            await initialize_database(self.context.engine)
        else:
            await check_catalog_database(self.context.engine)

    async def shutdown(self):
        await self.context.engine.dispose()

    @property
    def writable(self):
        return bool(self.context.writable_storage)

    def __repr__(self):
        return f"<{type(self).__name__} /{'/'.join(self.segments)}>"

    async def __aiter__(self):
        statement = select(orm.Node.key).filter(orm.Node.ancestors == self.segments)
        for condition in self.conditions:
            statement = statement.filter(condition)
        async with self.context.session() as db:
            return (
                (await db.execute(statement.order_by(*self.order_by_clauses)))
                .scalars()
                .all()
            )
        statement = select(orm.Node.key).filter(orm.Node.ancestors == self.segments)
        async with self.context.session() as db:
            return (await db.execute(statement)).scalar().all()

    @property
    def data_sources(self):
        return [DataSource.from_orm(ds) for ds in self.node.data_sources or []]

    def structure(self):
        if self.data_sources:
            assert len(self.data_sources) == 1  # more not yet implemented
            return self.data_sources[0].structure
        return None

    async def async_len(self):
        statement = select(func.count(orm.Node.key)).filter(
            orm.Node.ancestors == self.segments
        )
        for condition in self.conditions:
            statement = statement.filter(condition)
        async with self.context.session() as db:
            return (await db.execute(statement)).scalar_one()

    async def lookup_adapter(
        self, segments
    ):  # TODO: Accept filter for predicate-pushdown.
        if not segments:
            return self
        *ancestors, key = segments
        if self.conditions and len(segments) > 1:
            # There are some conditions (i.e. WHERE clauses) applied to
            # this node, either via user search queries or via access
            # control policy queries. Look up first the _direct_ child of this
            # node, if it exists within the filtered results.
            first_level = await self.lookup_adapter(segments[:1])
            if first_level is None:
                return None
            # Now proceed to traverse further down the tree, if needed.
            # Search queries and access controls apply only at the top level.
            assert not first_level.conditions
            return await first_level.lookup_adapter(segments[1:])
        statement = (
            select(orm.Node)
            .filter(orm.Node.ancestors == self.segments + ancestors)
            .options(
                selectinload(orm.Node.data_sources).selectinload(
                    orm.DataSource.structure
                )
            )
        )
        for condition in self.conditions:
            statement = statement.filter(condition)
        async with self.context.session() as db:
            node = (await db.execute(statement.filter(orm.Node.key == key))).scalar()
        if node is None:
            # Maybe the node does not exist, or maybe we have jumped _inside_ a file
            # whose internal contents are not indexed.

            # TODO As a performance optimization, do binary search or use some
            # tricky JSON query to find where database stops and (for example)
            # HDF5 file begins.

            for i in range(len(segments)):
                catalog_adapter = await self.lookup_adapter(segments[:i])
                if catalog_adapter.data_sources:
                    adapter = await catalog_adapter.get_adapter()
                    for segment in segments[i:]:
                        adapter = await anyio.to_thread.run_sync(adapter.get, segment)
                        if adapter is None:
                            break
                    return adapter
            return None
        return STRUCTURES[node.structure_family](
            self.context, node, access_policy=self.access_policy
        )

    async def get_adapter(self):
        num_data_sources = len(self.data_sources)
        if num_data_sources > 1:
            raise NotImplementedError
        if num_data_sources == 1:
            (data_source,) = self.data_sources
            try:
                adapter_factory = self.context.adapters_by_mimetype[
                    data_source.mimetype
                ]
            except KeyError:
                raise RuntimeError(
                    f"Server configuration has no adapter for mimetype {data_source.mimetype!r}"
                )
            parameters = collections.defaultdict(list)
            for asset in data_source.assets:
                if asset.parameter is None:
                    continue
                scheme = urlparse(asset.data_uri).scheme
                if scheme != "file":
                    raise NotImplementedError(
                        f"Only 'file://...' scheme URLs are currently supported, not {asset.data_uri}"
                    )
                if scheme == "file":
                    # Protect against misbehaving clients reading from unintended
                    # parts of the filesystem.
                    asset_path = path_from_uri(asset.data_uri)
                    for readable_storage in self.context.readable_storage:
                        if Path(
                            os.path.commonpath(
                                [path_from_uri(readable_storage), asset_path]
                            )
                        ) == path_from_uri(readable_storage):
                            break
                    else:
                        raise RuntimeError(
                            f"Refusing to serve {asset.data_uri} because it is outside "
                            "the readable storage area for this server."
                        )
                if asset.num is None:
                    parameters[asset.parameter] = asset.data_uri
                else:
                    parameters[asset.parameter].append(asset.data_uri)
            adapter_kwargs = dict(parameters)
            adapter_kwargs.update(data_source.parameters)
            adapter_kwargs["specs"] = self.node.specs
            adapter_kwargs["metadata"] = self.node.metadata_
            adapter_kwargs["structure"] = data_source.structure
            adapter_kwargs["access_policy"] = self.access_policy
            adapter = await anyio.to_thread.run_sync(
                partial(adapter_factory, **adapter_kwargs)
            )
            for query in self.queries:
                adapter = adapter.search(query)
            return adapter
        else:  # num_data_sources == 0
            assert False

    def new_variation(
        self,
        *args,
        sorting=UNCHANGED,
        conditions=UNCHANGED,
        queries=UNCHANGED,
        # must_revalidate=UNCHANGED,
        **kwargs,
    ):
        if sorting is UNCHANGED:
            sorting = self.sorting
        # if must_revalidate is UNCHANGED:
        #     must_revalidate = self.must_revalidate
        if conditions is UNCHANGED:
            conditions = self.conditions
        if queries is UNCHANGED:
            queries = self.queries
        return type(self)(
            self.context,
            node=self.node,
            conditions=conditions,
            sorting=sorting,
            # access_policy=self.access_policy,
            # entries_stale_after=self.entries_stale_after,
            # metadata_stale_after=self.entries_stale_after,
            # must_revalidate=must_revalidate,
            **kwargs,
        )

    def search(self, query):
        if self.data_sources:
            # Stand queries, which will be passed down to the adapter
            # when / if get_adapter() is called.
            self.queries.append(query)
            return self
        return self.query_registry(query, self)

    def sort(self, sorting):
        return self.new_variation(sorting=sorting)

    async def get_distinct(self, metadata, structure_families, specs, counts):
        if self.data_sources:
            return (await self.get_adapter()).get_disinct(
                metadata, structure_families, specs, counts
            )
        data = {}

        async with self.context.session() as db:
            if metadata:
                data["metadata"] = {}
                for key in metadata:
                    clause = orm.Node.metadata_
                    for segment in key.split("."):
                        clause = clause[segment]
                    if counts:
                        columns = (clause, func.count(clause))
                    else:
                        columns = (clause,)
                    statement = select(*columns).group_by(clause)
                    for condition in self.conditions:
                        statement = statement.filter(condition)
                    results = (await db.execute(statement)).all()
                    data["metadata"][key] = format_distinct_result(results, counts)

            if structure_families:
                if counts:
                    columns = (
                        orm.Node.structure_family,
                        func.count(orm.Node.structure_family),
                    )
                else:
                    columns = (orm.Node.structure_family,)
                statement = select(*columns).group_by(orm.Node.structure_family)
                for condition in self.conditions:
                    statement = statement.filter(condition)
                results = (await db.execute(statement)).all()

                data["structure_families"] = format_distinct_result(results, counts)

            if specs:
                if counts:
                    columns = (orm.Node.specs, func.count(orm.Node.specs))
                else:
                    columns = (orm.Node.specs,)
                statement = select(*columns).group_by(orm.Node.specs)
                for condition in self.conditions:
                    statement = statement.filter(condition)
                results = (await db.execute(statement)).all()

                data["specs"] = format_distinct_result(results, counts)

        return data

    async def create_node(
        self,
        structure_family,
        metadata,
        key=None,
        specs=None,
        data_sources=None,
    ):
        key = key or self.context.key_maker()
        data_sources = data_sources or []
        node = orm.Node(
            key=key,
            ancestors=self.segments,
            metadata_=metadata,
            structure_family=structure_family,
            specs=[s.dict() for s in specs or []],
        )
        async with self.context.session() as db:
            # TODO Consider using nested transitions to ensure that
            # both the node is created (name not already taken)
            # and the directory/file is created---or neither are.
            try:
                db.add(node)
                await db.commit()
            except IntegrityError as exc:
                UNIQUE_CONSTRAINT_FAILED = "gkpj"
                if exc.code == UNIQUE_CONSTRAINT_FAILED:
                    await db.rollback()
                    raise Collision(f"/{'/'.join(self.segments + [key])}")
                raise
            await db.refresh(node)
            for data_source in data_sources:
                if data_source.management != Management.external:
                    if structure_family == StructureFamily.container:
                        raise NotImplementedError(structure_family)
                    data_source.mimetype = DEFAULT_CREATION_MIMETYPE[structure_family]
                    data_source.parameters = {}
                    data_uri = str(self.context.writable_storage) + "".join(
                        f"/{quote_plus(segment)}" for segment in (self.segments + [key])
                    )
                    init_storage = DEFAULT_INIT_STORAGE[structure_family]
                    assets = await ensure_awaitable(
                        init_storage, data_uri, data_source.structure
                    )
                    data_source.assets.extend(assets)
                if data_source.structure is None:
                    structure_id = None
                else:
                    # Obtain and hash the canonical (RFC 8785) representation of
                    # the JSON structure.
                    structure = _prepare_structure(
                        structure_family, data_source.structure
                    )
                    structure_id = compute_structure_id(structure)
                    # The only way to do "insert if does not exist" i.e. ON CONFLICT
                    # is to invoke dialect-specific insert.
                    if self.context.engine.dialect.name == "sqlite":
                        from sqlalchemy.dialects.sqlite import insert
                    elif self.context.engine.dialect.name == "postgresql":
                        from sqlalchemy.dialects.postgresql import insert
                    else:
                        assert False  # future-proofing
                    statement = (
                        insert(orm.Structure).values(
                            id=structure_id,
                            structure=structure,
                        )
                    ).on_conflict_do_nothing(index_elements=["id"])
                    await db.execute(statement)
                data_source_orm = orm.DataSource(
                    mimetype=data_source.mimetype,
                    management=data_source.management,
                    parameters=data_source.parameters,
                    structure_id=structure_id,
                )
                node.data_sources.append(data_source_orm)
                # await db.flush(data_source_orm)
                for asset in data_source.assets:
                    asset_orm = orm.Asset(
                        data_uri=asset.data_uri,
                        is_directory=asset.is_directory,
                    )
                    assoc_orm = orm.DataSourceAssetAssociation(
                        asset=asset_orm,
                        parameter=asset.parameter,
                        num=asset.num,
                    )
                    data_source_orm.asset_associations.append(assoc_orm)
            db.add(node)
            await db.commit()
            await db.refresh(node)
            # Load with DataSources each DataSource's Structure.
            refreshed_node = (
                await db.execute(
                    select(orm.Node)
                    .filter(orm.Node.id == node.id)
                    .options(
                        selectinload(orm.Node.data_sources).selectinload(
                            orm.DataSource.structure
                        )
                    )
                )
            ).scalar()
            return key, type(self)(
                self.context, refreshed_node, access_policy=self.access_policy
            )

    # async def patch_node(datasources=None):
    #     ...

    async def revisions(self, offset, limit):
        async with self.context.session() as db:
            revision_orms = (
                await db.execute(
                    select(orm.Revision)
                    .where(orm.Revision.node_id == self.node.id)
                    .offset(offset)
                    .limit(limit)
                )
            ).all()
            return [Revision.from_orm(o[0]) for o in revision_orms]

    async def delete(self):
        """
        Delete a single Node.

        Any DataSources belonging to this Node and any Assets associated (only) with
        those DataSources will also be deleted.
        """
        async with self.context.session() as db:
            is_child = orm.Node.ancestors == self.ancestors + [self.key]
            num_children = (
                await db.execute(select(func.count(orm.Node.key)).where(is_child))
            ).scalar()
            if num_children:
                raise Conflicts(
                    "Cannot delete container that is not empty. Delete contents first."
                )
            for data_source in self.data_sources:
                if data_source.management != Management.external:
                    # TODO Handle case where the same Asset is associated
                    # with multiple DataSources. This is not possible yet
                    # but it is expected to become possible in the future.
                    for asset in data_source.assets:
                        delete_asset(asset.data_uri, asset.is_directory)
                        await db.execute(
                            delete(orm.Asset).where(orm.Asset.id == asset.id)
                        )
            result = await db.execute(
                delete(orm.Node).where(orm.Node.id == self.node.id)
            )
            if result.rowcount == 0:
                # TODO Abstract this from FastAPI?
                raise HTTPException(
                    status_code=404,
                    detail=f"No node {self.node.id}",
                )
            assert (
                result.rowcount == 1
            ), f"Deletion would affect {result.rowcount} rows; rolling back"
            await db.commit()

    async def delete_tree(self, external_only=True):
        """
        Delete a Node and of the Nodes beneath it in the tree.

        That is, delete all Nodes that have this Node as an ancestor, any number
        of "generators" up.

        Any DataSources belonging to those Nodes and any Assets associated (only) with
        those DataSources will also be deleted.
        """
        conditions = []
        segments = self.ancestors + [self.key]
        for generation in range(len(segments)):
            conditions.append(orm.Node.ancestors[generation] == segments[0])
        async with self.context.session() as db:
            if external_only:
                count_int_asset_statement = select(
                    func.count(orm.Asset.data_uri)
                ).filter(
                    orm.Asset.data_sources.any(
                        orm.DataSource.management != Management.external
                    )
                )
                for condition in conditions:
                    count_int_asset_statement.filter(condition)
                count_int_assets = (
                    await db.execute(count_int_asset_statement)
                ).scalar()
                if count_int_assets > 0:
                    raise WouldDeleteData(
                        "Some items in this tree are internally managed. "
                        "Delete the records will also delete the underlying data files. "
                        "If you want to delete them, pass external_only=False."
                    )
            else:
                sel_int_asset_statement = select(
                    orm.Asset.data_uri, orm.Asset.is_directory
                ).filter(
                    orm.Asset.data_sources.any(
                        orm.DataSource.management != Management.external
                    )
                )
                for condition in conditions:
                    sel_int_asset_statement.filter(condition)
                int_assets = (await db.execute(sel_int_asset_statement)).all()
                for data_uri, is_directory in int_assets:
                    delete_asset(data_uri, is_directory)
            # TODO Deal with Assets belonging to multiple DataSources.
            del_asset_statement = delete(orm.Asset)
            for condition in conditions:
                del_asset_statement.filter(condition)
            await db.execute(del_asset_statement)
            del_node_statement = delete(orm.Node)
            for condition in conditions:
                del_node_statement.filter(condition)
            result = await db.execute(del_node_statement)
            await db.commit()
        return result.rowcount

    async def delete_revision(self, number):
        async with self.context.session() as db:
            result = await db.execute(
                delete(orm.Revision)
                .where(orm.Revision.node_id == self.node.id)
                .where(orm.Revision.revision_number == number)
            )
            if result.rowcount == 0:
                # TODO Abstract this from FastAPI?
                raise HTTPException(
                    status_code=404,
                    detail=f"No revision {number} for node {self.node.id}",
                )
            assert (
                result.rowcount == 1
            ), f"Deletion would affect {result.rowcount} rows; rolling back"
            await db.commit()

    async def update_metadata(self, metadata=None, specs=None):
        values = {}
        if metadata is not None:
            # Trailing underscore in 'metadata_' avoids collision with
            # SQLAlchemy reserved word 'metadata'.
            values["metadata_"] = metadata
        if specs is not None:
            values["specs"] = [s.dict() for s in specs]
        async with self.context.session() as db:
            current = (
                await db.execute(select(orm.Node).where(orm.Node.id == self.node.id))
            ).scalar_one()
            next_revision_number = 1 + (
                (
                    await db.execute(
                        select(func.max(orm.Revision.revision_number)).where(
                            orm.Revision.node_id == self.node.id
                        )
                    )
                ).scalar_one()
                or 0
            )
            revision = orm.Revision(
                # Trailing underscore in 'metadata_' avoids collision with
                # SQLAlchemy reserved word 'metadata'.
                metadata_=current.metadata_,
                specs=current.specs,
                node_id=current.id,
                revision_number=next_revision_number,
            )
            db.add(revision)
            await db.execute(
                update(orm.Node).where(orm.Node.id == self.node.id).values(**values)
            )
            await db.commit()


class CatalogContainerAdapter(CatalogNodeAdapter):
    async def keys_range(self, offset, limit):
        if self.data_sources:
            return (await self.get_adapter()).keys()[
                offset : (offset + limit) if limit is not None else None  # noqa: E203
            ]
        statement = select(orm.Node.key).filter(orm.Node.ancestors == self.segments)
        for condition in self.conditions:
            statement = statement.filter(condition)
        async with self.context.session() as db:
            return (
                (
                    await db.execute(
                        statement.order_by(*self.order_by_clauses)
                        .offset(offset)
                        .limit(limit)
                    )
                )
                .scalars()
                .all()
            )

    async def items_range(self, offset, limit):
        if self.data_sources:
            return (await self.get_adapter()).items()[
                offset : (offset + limit) if limit is not None else None  # noqa: E203
            ]
        statement = select(orm.Node).filter(orm.Node.ancestors == self.segments)
        for condition in self.conditions:
            statement = statement.filter(condition)
        async with self.context.session() as db:
            nodes = (
                (
                    await db.execute(
                        statement.order_by(*self.order_by_clauses)
                        .offset(offset)
                        .limit(limit)
                    )
                )
                .scalars()
                .all()
            )
            return [
                (
                    node.key,
                    type(self)(self.context, node, access_policy=self.access_policy),
                )
                for node in nodes
            ]

    async def read(self, *args, **kwargs):
        if not self.data_sources:
            fields = kwargs.get("fields")
            if fields:
                return self.search(KeysFilter(fields))
            return self
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)


class CatalogArrayAdapter(CatalogNodeAdapter):
    async def read(self, *args, **kwargs):
        if not self.data_sources:
            fields = kwargs.get("fields")
            if fields:
                return self.search(KeysFilter(fields))
            return self
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)

    async def read_block(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).read_block, *args, **kwargs
        )

    async def write(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).write, *args, **kwargs)

    async def write_block(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).write_block, *args, **kwargs
        )


class CatalogAwkwardAdapter(CatalogNodeAdapter):
    async def read(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)

    async def read_buffers(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).read_buffers, *args, **kwargs
        )

    async def write(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).write, *args, **kwargs)


class CatalogSparseAdapter(CatalogArrayAdapter):
    pass


class CatalogTableAdapter(CatalogNodeAdapter):
    async def read(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)

    async def write(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).write, *args, **kwargs)

    async def read_partition(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).read_partition, *args, **kwargs
        )

    async def write_partition(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).write_partition, *args, **kwargs
        )


def delete_asset(data_uri, is_directory):
    url = urlparse(data_uri)
    if url.scheme == "file":
        path = path_from_uri(data_uri)
        if is_directory:
            shutil.rmtree(path)
        else:
            Path(path).unlink()
    else:
        raise NotImplementedError(url.scheme)


_STANDARD_SORT_KEYS = {
    "id": "key",
    # Maybe add things like structure_family here, but it's not clear what the
    # sort order would be.
}


def order_by_clauses(sorting):
    clauses = []
    default_sorting_direction = 1
    for key, direction in sorting:
        if key == "":
            default_sorting_direction = direction
            continue
            # TODO Really we should insist that if this is given, it is last,
            # because we always apply the default sorting last.
        if key in _STANDARD_SORT_KEYS:
            clause = getattr(orm.Node, _STANDARD_SORT_KEYS[key])
        else:
            clause = orm.Node.metadata_
            # This can be given bare like "color" or namedspaced like
            # "metadata.color" to avoid the possibility of a name collision
            # with the standard sort keys.
            if key.startswith("metadata."):
                key = key[len("metadata.") :]  # noqa: E203

            for segment in key.split("."):
                clause = clause[segment]
        if direction == -1:
            clause = clause.desc()
        clauses.append(clause)
    # Ensure deterministic ordering for all queries by sorting by
    # 'time_created' and then by 'id' last.
    for clause in [orm.Node.time_created, orm.Node.id]:
        if default_sorting_direction == -1:
            clause = clause.desc()
        clauses.append(clause)
    return clauses


_TYPE_CONVERSION_MAP = {
    int: "as_integer",
    float: "as_float",
    str: "as_string",
    bool: "as_boolean",
}


def _get_value(value, type):
    # This is used only for SQLite, to get the types right
    # so that that index is used. There is probably a
    # cleaner way to handle this.
    # Study https://gist.github.com/brthor/e3d23ae549ee53cdea56d72d39ad1288
    # which may or may not be relevant anymore.
    return getattr(value, _TYPE_CONVERSION_MAP[type])()


def _prepare_structure(structure_family, structure):
    "Convert from pydantic model to dict."
    if structure is None:
        return None
    return structure.dict()


def binary_op(query, tree, operation):
    dialect_name = tree.engine.url.get_dialect().name
    keys = query.key.split(".")
    attr = orm.Node.metadata_[keys]
    if dialect_name == "sqlite":
        condition = operation(_get_value(attr, type(query.value)), query.value)
    # specific case where GIN optomized index can be used to speed up POSTGRES equals queries
    elif (dialect_name == "postgresql") and (operation == operator.eq):
        condition = orm.Node.metadata_.op("@>")(
            type_coerce(
                key_array_to_json(keys, query.value),
                orm.Node.metadata_.type,
            )
        )
    else:
        condition = operation(attr, type_coerce(query.value, orm.Node.metadata_.type))
    return tree.new_variation(conditions=tree.conditions + [condition])


def comparison(query, tree):
    OPERATORS = {
        Operator.lt: operator.lt,
        Operator.le: operator.le,
        Operator.gt: operator.gt,
        Operator.ge: operator.ge,
    }
    return binary_op(query, tree, OPERATORS[query.operator])


def contains(query, tree):
    attr = orm.Node.metadata_[query.key.split(".")]
    condition = _get_value(attr, type(query.value)).contains(query.value)
    return tree.new_variation(conditions=tree.conditions + [condition])


def full_text(query, tree):
    dialect_name = tree.engine.url.get_dialect().name
    if dialect_name == "sqlite":
        raise UnsupportedQueryType("full_text")
    elif dialect_name == "postgresql":
        tsvector = func.jsonb_to_tsvector(
            cast("simple", REGCONFIG), orm.Node.metadata_, cast(["string"], JSONB)
        )
        condition = tsvector.op("@@")(func.to_tsquery("simple", query.text))
    else:
        raise UnsupportedQueryType("full_text")
    return tree.new_variation(conditions=tree.conditions + [condition])


def specs(query, tree):
    dialect_name = tree.engine.url.get_dialect().name
    conditions = []
    attr = orm.Node.specs
    if dialect_name == "sqlite":
        # Construct the conditions for includes
        for i, name in enumerate(query.include):
            conditions.append(attr.like(f'%{{"name":"{name}",%'))
        # Construct the conditions for excludes
        for i, name in enumerate(query.exclude):
            conditions.append(not_(attr.like(f'%{{"name":"{name}",%')))
    elif dialect_name == "postgresql":
        if query.include:
            conditions.append(attr.op("@>")(specs_array_to_json(query.include)))
        if query.exclude:
            conditions.append(not_(attr.op("@>")(specs_array_to_json(query.exclude))))
    else:
        raise UnsupportedQueryType("specs")
    return tree.new_variation(conditions=tree.conditions + conditions)


def in_or_not_in(query, tree, method):
    dialect_name = tree.engine.url.get_dialect().name
    keys = query.key.split(".")
    attr = orm.Node.metadata_[keys]
    if dialect_name == "sqlite":
        condition = getattr(_get_value(attr, type(query.value[0])), method)(query.value)
    elif dialect_name == "postgresql":
        # Engage btree_gin index with @> operator
        if method == "in_":
            condition = or_(
                *(
                    orm.Node.metadata_.op("@>")(key_array_to_json(keys, item))
                    for item in query.value
                )
            )
        elif method == "not_in":
            condition = not_(
                or_(
                    *(
                        orm.Node.metadata_.op("@>")(key_array_to_json(keys, item))
                        for item in query.value
                    )
                )
            )
        else:
            raise UnsupportedQueryType("NotIn")
    else:
        raise UnsupportedQueryType("In/NotIn")
    return tree.new_variation(conditions=tree.conditions + [condition])


def keys_filter(query, tree):
    condition = orm.Node.key.in_(query.keys)
    return tree.new_variation(conditions=tree.conditions + [condition])


def structure_family(query, tree):
    condition = orm.Node.structure_family == query.value
    return tree.new_variation(conditions=tree.conditions + [condition])


CatalogNodeAdapter.register_query(Eq, partial(binary_op, operation=operator.eq))
CatalogNodeAdapter.register_query(NotEq, partial(binary_op, operation=operator.ne))
CatalogNodeAdapter.register_query(Comparison, comparison)
CatalogNodeAdapter.register_query(Contains, contains)
CatalogNodeAdapter.register_query(In, partial(in_or_not_in, method="in_"))
CatalogNodeAdapter.register_query(NotIn, partial(in_or_not_in, method="not_in"))
CatalogNodeAdapter.register_query(KeysFilter, keys_filter)
CatalogNodeAdapter.register_query(StructureFamilyQuery, structure_family)
CatalogNodeAdapter.register_query(SpecsQuery, specs)
CatalogNodeAdapter.register_query(FullText, full_text)
# TODO: Regex


def in_memory(
    *,
    metadata=None,
    specs=None,
    access_policy=None,
    writable_storage=None,
    readable_storage=None,
    echo=DEFAULT_ECHO,
    adapters_by_mimetype=None,
):
    uri = "sqlite+aiosqlite:///:memory:"
    return from_uri(
        uri=uri,
        metadata=metadata,
        specs=specs,
        access_policy=access_policy,
        writable_storage=writable_storage,
        readable_storage=readable_storage,
        echo=echo,
        adapters_by_mimetype=adapters_by_mimetype,
    )


def from_uri(
    uri,
    *,
    metadata=None,
    specs=None,
    access_policy=None,
    writable_storage=None,
    readable_storage=None,
    init_if_not_exists=False,
    echo=DEFAULT_ECHO,
    adapters_by_mimetype=None,
):
    uri = str(uri)
    if init_if_not_exists:
        # The alembic stamping can only be does synchronously.
        # The cleanest option available is to start a subprocess
        # because SQLite is allergic to threads.
        import subprocess

        # TODO Check if catalog exists.
        subprocess.run(
            [sys.executable, "-m", "tiled", "catalog", "init", "--if-not-exists", uri],
            capture_output=True,
            check=True,
        )
    if not SCHEME_PATTERN.match(uri):
        # Interpret URI as filepath.
        uri = f"sqlite+aiosqlite:///{uri}"

    engine = create_async_engine(uri, echo=echo, json_serializer=json_serializer)
    if engine.dialect.name == "sqlite":
        event.listens_for(engine.sync_engine, "connect")(_set_sqlite_pragma)
    return CatalogContainerAdapter(
        Context(engine, writable_storage, readable_storage, adapters_by_mimetype),
        RootNode(metadata, specs, access_policy),
        access_policy=access_policy,
    )


def _set_sqlite_pragma(conn, record):
    cursor = conn.cursor()
    # https://docs.sqlalchemy.org/en/13/dialects/sqlite.html#foreign-key-support
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def format_distinct_result(results, counts):
    if counts:
        formatted_result = [
            {"value": value, "count": count} for value, count in results
        ]
    else:
        formatted_result = [{"value": value} for value, in results]
    return formatted_result


class WouldDeleteData(RuntimeError):
    pass


class Collision(Conflicts):
    pass


def json_serializer(obj):
    "The PostgreSQL JSON serializer requires str, not bytes."
    return safe_json_dump(obj).decode()


def key_array_to_json(keys, value):
    """Take JSON accessor information as an array of keys and value

    Parameters
    ----------
    keys : iterable
        An array of keys to be created in the object.
    value : string
        Value assigned to the final key.

    Returns
    -------
    json
        JSON object for use in postgresql queries.

    Examples
    --------
    >>> key_array_to_json(['x','y','z'], 1)
    {'x': {'y': {'z': 1}}
    """
    return {keys[0]: reduce(lambda x, y: {y: x}, keys[1:][::-1], value)}


def specs_array_to_json(specs):
    """Take array of Specs strings and convert them to a `penguin` @> friendly array
    Assume constructed array will feature keys called "name"

    Parameters
    ----------
    specs : iterable
        An array of specs strings to be searched for.

    Returns
    -------
    json
        JSON object for use in postgresql queries.

    Examples
    --------
    >>> specs_array_to_json(['foo','bar'])
    [{"name":"foo"},{"name":"bar"}]
    """
    return [{"name": spec} for spec in specs]


STRUCTURES = {
    StructureFamily.container: CatalogContainerAdapter,
    StructureFamily.array: CatalogArrayAdapter,
    StructureFamily.awkward: CatalogAwkwardAdapter,
    StructureFamily.table: CatalogTableAdapter,
    StructureFamily.sparse: CatalogSparseAdapter,
}



from pathlib import Path

import yaml

from ..client import from_profile
from ..config import parse_configs
from ..profiles import load_profiles, paths
from ..server.app import build_app_from_config

here = Path(__file__).parent.absolute()


def test_config_imports_custom_python_module():
    "Configs can import from Python modules located in their same directory."
    config_path = here / ".." / ".." / "example_configs" / "custom_export_formats"
    parsed_config = parse_configs(config_path)
    build_app_from_config(parsed_config, source_filepath=config_path)


def test_direct_profile(tmpdir):
    profile_content = {
        "test": {"direct": {"trees": [{"path": "/", "tree": "custom_module:tree"}]}}
    }
    with open(tmpdir / "example.yml", "w") as yaml_file:
        yaml_file.write(yaml.dump(profile_content))
    with open(tmpdir / "custom_module.py", "w") as py_file:
        py_file.write(
            "from tiled.adapters.mapping import MapAdapter; tree = MapAdapter({})"
        )
    profile_dir = Path(tmpdir)
    try:
        paths.append(profile_dir)
        load_profiles.cache_clear()
        from_profile("test")
    finally:
        paths.remove(profile_dir)



import io
import json

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..server.app import build_app_from_config
from .utils import enter_password, fail_with_status_code

arr = numpy.ones((5, 5))
arr_ad = ArrayAdapter.from_array(arr)


def tree_a(access_policy=None):
    return MapAdapter({"A1": arr_ad, "A2": arr_ad}, access_policy=access_policy)


def tree_b(access_policy=None):
    return MapAdapter({"B1": arr_ad, "B2": arr_ad}, access_policy=access_policy)


@pytest.fixture(scope="module")
def context(tmpdir_module):
    config = {
        "authentication": {
            "allow_anonymous_access": True,
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {
                            "alice": "secret1",
                            "bob": "secret2",
                            "admin": "admin",
                        }
                    },
                }
            ],
        },
        "database": {
            "uri": "sqlite+aiosqlite://",  # in-memory
        },
        "access_control": {
            "access_policy": "tiled.access_policies:SimpleAccessPolicy",
            "args": {
                "access_lists": {"alice": ["a", "c", "d", "e"]},
                "provider": "toy",
                "admins": ["admin"],
                "public": ["f"],
            },
        },
        "trees": [
            {
                "tree": f"{__name__}:tree_a",
                "path": "/a",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": ["A2"],
                            # This should have no effect because bob
                            # cannot access the parent node.
                            "bob": ["A1", "A2"],
                        },
                        "admins": ["admin"],
                    },
                },
            },
            {"tree": f"{__name__}:tree_b", "path": "/b", "access_policy": None},
            {
                "tree": "tiled.catalog:in_memory",
                "args": {"writable_storage": tmpdir_module / "c"},
                "path": "/c",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": "tiled.access_policies:ALL_ACCESS",
                        },
                        "admins": ["admin"],
                    },
                },
            },
            {
                "tree": "tiled.catalog:in_memory",
                "args": {"writable_storage": tmpdir_module / "d"},
                "path": "/d",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": "tiled.access_policies:ALL_ACCESS",
                        },
                        "admins": ["admin"],
                        # Block writing.
                        "scopes": ["read:metadata", "read:data"],
                    },
                },
            },
            {
                "tree": "tiled.catalog:in_memory",
                "args": {"writable_storage": tmpdir_module / "e"},
                "path": "/e",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": "tiled.access_policies:ALL_ACCESS",
                        },
                        "admins": ["admin"],
                        # Block creation.
                        "scopes": [
                            "read:metadata",
                            "read:data",
                            "write:metadata",
                            "write:data",
                        ],
                    },
                },
            },
            {"tree": ArrayAdapter.from_array(arr), "path": "/f"},
        ],
    }
    app = build_app_from_config(config)
    with Context.from_app(app) as context:
        with enter_password("admin"):
            admin_client = from_context(context, username="admin")
            for k in ["c", "d", "e"]:
                admin_client[k].write_array(arr, key="A1")
                admin_client[k].write_array(arr, key="A2")
                admin_client[k].write_array(arr, key="x")
        yield context


def test_top_level_access_control(context, enter_password):
    with enter_password("secret1"):
        alice_client = from_context(context, username="alice")
    assert "a" in alice_client
    assert "A2" in alice_client["a"]
    assert "A1" not in alice_client["a"]
    assert "b" not in alice_client
    alice_client["a"]["A2"]
    with pytest.raises(KeyError):
        alice_client["b"]

    with enter_password("secret2"):
        bob_client = from_context(context, username="bob")
    assert not list(bob_client)
    with pytest.raises(KeyError):
        bob_client["a"]
    with pytest.raises(KeyError):
        bob_client["b"]
    alice_client.logout()
    bob_client.logout()


def test_access_control_with_api_key_auth(context, enter_password):
    # Log in, create an API key, log out.
    with enter_password("secret1"):
        context.authenticate(username="alice")
    key_info = context.create_api_key()
    context.logout()

    try:
        # Use API key auth while exercising the access control code.
        context.api_key = key_info["secret"]
        client = from_context(context)
        client["a"]["A2"]
    finally:
        # Clean up Context, which is a module-scopae fixture shared with other tests.
        context.api_key = None


def test_node_export(enter_password, context):
    "Exporting a node should include only the children we can see."
    with enter_password("secret1"):
        alice_client = from_context(context, username="alice")
    buffer = io.BytesIO()
    alice_client.export(buffer, format="application/json")
    alice_client.logout()
    buffer.seek(0)
    exported_dict = json.loads(buffer.read())
    assert "a" in exported_dict["contents"]
    assert "A2" in exported_dict["contents"]["a"]["contents"]
    assert "A1" not in exported_dict["contents"]["a"]["contents"]
    assert "b" not in exported_dict
    exported_dict["contents"]["a"]["contents"]["A2"]


def test_create_and_update_allowed(enter_password, context):
    with enter_password("secret1"):
        alice_client = from_context(context, username="alice")

    # Update
    alice_client["c"]["x"].metadata
    alice_client["c"]["x"].update_metadata(metadata={"added_key": 3})
    assert alice_client["c"]["x"].metadata["added_key"] == 3

    # Create
    alice_client["c"].write_array([1, 2, 3])
    alice_client.logout()


def test_writing_blocked_by_access_policy(enter_password, context):
    with enter_password("secret1"):
        alice_client = from_context(context, username="alice")
    alice_client["d"]["x"].metadata
    with fail_with_status_code(403):
        alice_client["d"]["x"].update_metadata(metadata={"added_key": 3})
    alice_client.logout()


def test_create_blocked_by_access_policy(enter_password, context):
    with enter_password("secret1"):
        alice_client = from_context(context, username="alice")
    with fail_with_status_code(403):
        alice_client["e"].write_array([1, 2, 3])
    alice_client.logout()


def test_public_access(context):
    public_client = from_context(context)
    for key in ["a", "b", "c", "d", "e"]:
        assert key not in public_client
    public_client["f"].read()
    with pytest.raises(KeyError):
        public_client["a", "A1"]



"""
This module handles client configuration.

See config.py for server configuration.

It contains several functions that are factored to facilitate testing,
but the user-facing functionality is striaghtforward.
"""
import collections
import collections.abc
import os
import sys
import warnings
from functools import lru_cache
from pathlib import Path

import appdirs
import jsonschema

from .utils import parse

__all__ = [
    "list_profiles",
    "load_profiles",
    "paths",
    "create_profile",
    "delete_profile",
    "set_default_profile_name",
    "get_default_profile_name",
]


@lru_cache(maxsize=1)
def schema():
    "Load the schema for profiles."
    import yaml

    here = Path(__file__).parent.absolute()
    schema_path = here / "config_schemas" / "client_profiles.yml"
    with open(schema_path, "r") as file:
        return yaml.safe_load(file)


# Some items in the search path is system-dependent, and others are hard-coded.
# Paths later in the list ("closer" to the user) have higher precedence.
_all_paths = [
    Path(
        os.getenv("TILED_SITE_PROFILES", Path("/etc/tiled/profiles"))
    ),  # hard-coded system path
    Path(
        os.getenv(
            "TILED_SITE_PROFILES", Path(appdirs.site_config_dir("tiled"), "profiles")
        )
    ),  # XDG-compliant system path
    Path(sys.prefix, "etc", "tiled", "profiles"),  # environment
    Path(
        os.getenv("TILED_PROFILES", Path.home() / ".config/tiled/profiles")
    ),  # hard-coded user path
    Path(
        os.getenv("TILED_PROFILES", Path(appdirs.user_config_dir("tiled"), "profiles"))
    ),  # system-dependent user path
]
# Remove duplicates (i.e. if XDG and hard-coded are the same on this system).
_seen = set()
paths = [x for x in _all_paths if not (x in _seen or _seen.add(x))]
del _seen


def gather_profiles(paths, strict=True):
    """
    For each path in paths, return a dict mapping filepath to content.
    """
    levels = []
    for path in paths:
        filepath_to_content = {}
        if path.is_dir():
            for filename in path.iterdir():
                filepath = path / filename
                # Ignore hidden files and .py files.
                if (
                    filename.name.startswith(".")
                    or filename.suffix == ".py"
                    or filename.name == "__pycache__"
                ):
                    continue
                try:
                    with open(filepath) as file:
                        content = parse(file)
                        if content is None:
                            raise ProfileError(f"File {filepath!s} is empty.")
                        if not isinstance(content, collections.abc.Mapping):
                            raise ProfileError(
                                f"File {filepath!s} does not have the expected structure."
                            )
                        for profile_name, profile_content in content.items():
                            try:
                                jsonschema.validate(
                                    instance=profile_content, schema=schema()
                                )
                            except jsonschema.ValidationError as validation_err:
                                original_msg = validation_err.args[0]
                                raise ProfileError(
                                    f"ValidationError while parsing profile {profile_name} "
                                    f"in file {filepath!s}: {original_msg}"
                                ) from validation_err
                except Exception as err:
                    if strict:
                        raise
                    else:
                        warnings.warn(
                            f"Skipping {filepath!s}. Failed to parse with error: {err!r}."
                        )
                        continue
                if not isinstance(content, collections.abc.Mapping):
                    if strict:
                        raise
                    else:
                        warnings.warn(
                            f"Skipping {filepath!r}. Content has invalid structure (not a mapping)."
                        )
                filepath_to_content[filepath] = content
        levels.append(filepath_to_content)
    return levels


def resolve_precedence(levels):
    """
    Given a list of mappings (filename-to-content), resolve precedence.

    In the event of irresolvable collisions, drop the offenders and warn.

    The result is a mapping from profile name to (filepath, content).
    """
    # Map each profile_name to the file(s) that define a profile with that name.
    # This is used to track collisions.
    profile_name_to_filepaths_per_level = [
        collections.defaultdict(list) for _ in range(len(paths))
    ]
    for profile_name_to_filepaths, filepath_to_content in zip(
        profile_name_to_filepaths_per_level, levels
    ):
        for filepath, content in filepath_to_content.items():
            for profile_name in content:
                profile_name_to_filepaths[profile_name].append(filepath)
    combined = {}
    collisions = {}
    for profile_name_to_filepaths, filepath_to_content in zip(
        profile_name_to_filepaths_per_level, levels
    ):
        for profile_name, filepaths in profile_name_to_filepaths.items():
            # A profile name in this level resolves any collisions in the previous level.
            collisions.pop(profile_name, None)
            # Does more than one file *in this same level (directory) in the search path*
            # define a profile with the same name? If so, there is no sure way to decide
            # precedence so we will omit it and issue a warning, unless something later
            # in the search path overrides it.
            if len(filepaths) > 1:
                # Stash collision for possible warning below.
                collisions[profile_name] = filepaths
                # If a previous level defined this, remove it.
                combined.pop(profile_name, None)
            else:
                (filepath,) = filepaths
                combined[profile_name] = (
                    filepath,
                    filepath_to_content[filepath][profile_name],
                )
    MSG = """More than file in the same directory:

{filepaths}

defines a profile with the name {profile_name!r}.

The profile will be ommitted. Fix this by removing one of the duplicates"""
    for profile_name, filepaths in collisions.items():
        # Is this file in either the XDG user or hard-coded user directory
        # (which might or might not be the same directory)?
        if filepaths[0].is_relative_to(_all_paths[-1]) or filepaths[0].is_relative_to(
            _all_paths[-2]
        ):
            msg = (MSG + ".").format(
                filepaths="\n".join(map(str, filepaths)), profile_name=profile_name
            )
        else:
            msg = (
                MSG
                + (
                    " or by defining a profile with that name in a "
                    f"file in the user config directory {paths[-1]} "
                    "to override them."
                )
            ).format(
                filepaths="\n".join(map(str, filepaths)), profile_name=profile_name
            )
        warnings.warn(msg)
    return combined


@lru_cache(maxsize=1)
def load_profiles():
    """
    Return a mapping of profile_name to (source_path, content).

    The files are only actually read the first time this is called.
    Thereafter, the results are cached. To clear the cache and re-read,
    use load_profiles.cache_clear().

    The search path for the source files is available from Python as:

    >>> tiled.client.profiles.paths

    or from a CLI as:

    $ tiled profile paths

    See also

    $ tiled profile list
    $ tiled profile show PROFILE_NAME
    """
    levels = gather_profiles(paths, strict=False)
    profiles = resolve_precedence(levels)
    return profiles


def list_profiles():
    """
    Return a mapping of profile names to source filepath.

    The search path for the source files is available from Python as:

    >>> tiled.client.profiles.paths

    or from a CLI as:

    $ tiled profile paths

    See also

    $ tiled profile list
    $ tiled profile show PROFILE_NAME
    """
    levels = gather_profiles(paths, strict=False)
    profiles = resolve_precedence(levels)
    return {name: source_filepath for name, (source_filepath, _) in profiles.items()}


def _compose_profile(name, *, uri, verify):
    "Compose profile YAML."
    import yaml

    content = {name: {"uri": uri, "verify": verify}}
    return yaml.dump(content)


def create_profile(uri, name, verify=True, overwrite=False):
    """
    Create a new profile.

    This only includes the most commonly-used options. (More could be added.)
    """
    text = _compose_profile(name, uri=uri, verify=verify)
    filepath = paths[-1] / f"{name}.yml"
    filepath.parent.mkdir(parents=True, exist_ok=True)
    if overwrite:
        mode = "wt"
    else:
        mode = "xt"
    try:
        with open(filepath, mode) as file:
            file.write(text)
    except FileExistsError:
        raise ProfileExists(
            f"Profile named {name} already exists at {filepath}. "
            "Use overwite=True to overwrite it."
        )
    return filepath


def delete_profile(name):
    """
    Delete a profile by name.

    This will walk the search path, starting with the highest precedence
    directory, and delete only the first match it finds.
    """
    for path in paths:
        # All profiles created by create_profile have extension .yml but
        # a user-written one may have extension .yaml.
        for ext in {".yml", ".yaml"}:
            filepath = path / f"{name}{ext}"
            if filepath.exists():
                filepath.unlink()
                return filepath


def get_default_profile_name():
    """
    Return the name of the current default profile.
    """
    filepath = paths[-1].parent / "default_profile"
    try:
        return filepath.read_text()
    except FileNotFoundError:
        return None


def set_default_profile_name(name):
    filepath = paths[-1].parent / "default_profile"
    filepath.parent.mkdir(parents=True, exist_ok=True)
    if name is None:
        if filepath.exists():
            filepath.unlink()
        return
    if name not in list_profiles():
        raise ProfileNotFound(name)
    with open(filepath, "w") as file:
        file.write(name)


class ProfileNotFound(KeyError):
    pass


class ProfileError(ValueError):
    pass


class ProfileExists(ValueError):
    pass



import collections
import contextlib
import logging
import os

from ..utils import bytesize_repr

# By default, the token in the authentication header is redacted from the logs.
# Set this env var to 1 to show it for debugging purposes.
TILED_LOG_AUTH_TOKEN = int(os.getenv("TILED_LOG_AUTH_TOKEN", False))


class ClientLogRecord(logging.LogRecord):
    def getMessage(self):
        if hasattr(self, "request"):
            request = self.request
            if "content-length" in request.headers:
                size = f"({bytesize_repr(int(request.headers['content-length']))})"
            else:
                size = ""
            message = f"-> {size} {request.method} '{request.url}' " + " ".join(
                f"'{k}:{v}'" for k, v in request.headers.items() if k != "authorization"
            )
            # Handle the authorization header specially.
            # For debugging, it can be useful to show it so that the log message
            # can be copy/pasted and passed to httpie in a shell.
            # But for screen-sharing demos, it should be redacted.
            if TILED_LOG_AUTH_TOKEN:
                if "authorization" in request.headers:
                    message += f" 'authorization:{request.headers['authorization']}'"
            else:
                if "authorization" in request.headers:
                    scheme, _, param = request.headers["authorization"].partition(" ")
                    message += f" 'authorization:{scheme} [redacted]'"
        elif hasattr(self, "response"):
            response = self.response
            request = response.request
            if "content-length" in response.headers:
                size = f"({bytesize_repr(int(response.headers['content-length']))})"
            else:
                size = ""
            message = f"<- {size} {response.status_code} " + " ".join(
                f"{k}:{v}" for k, v in response.headers.items()
            )
        else:
            message = super().getMessage()
        return message


def patched_make_record(
    name,
    level,
    fn,
    lno,
    msg,
    args,
    exc_info,
    func=None,
    extra=None,
    sinfo=None,
):
    rv = ClientLogRecord(name, level, fn, lno, msg, args, exc_info, func, sinfo)
    if extra is not None:
        for key in extra:
            if (key in ["message", "asctime"]) or (key in rv.__dict__):
                raise KeyError("Attempt to overwrite %r in LogRecord" % key)
            rv.__dict__[key] = extra[key]
    return rv


logger = logging.getLogger("tiled.client")
# Monkey-patch our logger!
# The logging framework provides no way to look a custom record factory into
# the global logging manager. I tried several ways to avoid monkey-patching
# and this is the least bad. Notice that it only downloades the 'tiled.client'
# logger and will not affect the behavior of other loggers.
logger.makeRecord = patched_make_record
handler = logging.StreamHandler()
log_format = "%(asctime)s.%(msecs)03d %(message)s"

handler.setFormatter(logging.Formatter(log_format, datefmt="%H:%M:%S"))


def log_request(request):
    logger.debug("", extra={"request": request})


def log_response(response):
    logger.debug("", extra={"response": response})


def collect_request(request):
    if _history is not None:
        _history.requests.append(request)


def collect_response(response):
    if _history is not None:
        _history.responses.append(response)


def show_logs():
    """
    Log network traffic and interactions with the cache.

    This is just a convenience function that makes some Python logging configuration calls.
    """
    logger.setLevel("DEBUG")
    logger.addHandler(handler)


def hide_logs():
    """
    Undo show_logs().
    """
    logger.setLevel("WARNING")
    if handler in logger.handlers:
        logger.removeHandler(handler)


History = collections.namedtuple("History", "requests responses")
_history = None


@contextlib.contextmanager
def record_history():
    """
    Collect requests and responses.

    >>> with history() as t:
    ...     ...
    ...

    >>> t.requests
    [...]

    >>> t.responses
    [...]
    """
    global _history

    responses = []
    requests = []
    history = History(requests, responses)
    _history = history
    yield history
    _history = None



# This file will be (temporarily) included in the Python sys.path
# when config.yml is loaded by the Tiled server.
import io

from PIL import Image

from tiled.serialization.image_serializer_helpers import img_as_ubyte


def smiley_separated_variables(array, metadata):
    return "\n".join("🙂".join(str(number) for number in row) for row in array)


def to_jpeg(array, metadata):
    file = io.BytesIO()
    # PIL detail: ensure array has compatible data type before handing to PIL.
    prepared_array = img_as_ubyte(array)
    image = Image.fromarray(prepared_array)
    image.save(file, format="jpeg")
    return file.getbuffer()



import io

from ..media_type_registration import serialization_registry
from ..structures.core import StructureFamily
from ..utils import SerializationError, modules_available, safe_json_dump


async def walk(node, filter_for_access, pre=None):
    """
    Yield (key_path, value) where each value is an ArrayAdapter.

    As a succinct illustration (does not literally run):

    >>> list(walk{"a": {"b": 1, "c": {"d": 2}}})
    [
        (("a", "b"), 1),
        (("a", "c", "d"), 2),
    ]
    """
    pre = pre[:] if pre else []
    if node.structure_family != StructureFamily.array:
        if hasattr(node, "items_range"):
            for key, value in await filter_for_access(node).items_range(0, None):
                async for d in walk(value, filter_for_access, pre + [key]):
                    yield d
        else:
            for key, value in filter_for_access(node).items():
                async for d in walk(value, filter_for_access, pre + [key]):
                    yield d
    else:
        yield (pre, node)


if modules_available("h5py"):

    async def serialize_hdf5(node, metadata, filter_for_access):
        """
        Encode everything below this node as HDF5.

        Walk node. Encode all nodes an dataframes as Groups, all arrays and columns as Datasets.
        """
        import h5py

        buffer = io.BytesIO()
        root_node = node
        MSG = "Metadata contains types or structure that does not fit into HDF5."
        with h5py.File(buffer, mode="w") as file:
            try:
                file.attrs.update(metadata)
            except TypeError:
                raise SerializationError(MSG)
            async for key_path, array_adapter in walk(node, filter_for_access):
                group = file
                node = root_node
                for key in key_path[:-1]:
                    if hasattr(node, "lookup_adapter"):
                        node = await node.lookup_adapter([key])
                    else:
                        node = node[key]
                    if key in group:
                        group = group[key]
                    else:
                        group = group.create_group(key)
                        try:
                            group.attrs.update(node.metadata())
                        except TypeError:
                            raise SerializationError(MSG)
                data = array_adapter.read()
                dataset = group.create_dataset(key_path[-1], data=data)
                for k, v in array_adapter.metadata().items():
                    dataset.attrs.create(k, v)
        return buffer.getbuffer()

    serialization_registry.register(
        StructureFamily.container, "application/x-hdf5", serialize_hdf5
    )

if modules_available("orjson"):

    async def serialize_json(node, metadata, filter_for_access):
        "Export node to JSON, with each node having a 'contents' and 'metadata' sub-key."
        root_node = node
        to_serialize = {"contents": {}, "metadata": dict(root_node.metadata())}
        async for key_path, array_adapter in walk(node, filter_for_access):
            d = to_serialize["contents"]
            node = root_node
            for key in key_path:
                if hasattr(node, "lookup_adapter"):
                    node = await node.lookup_adapter([key])
                else:
                    node = node[key]
                if key not in d:
                    d[key] = {"contents": {}, "metadata": dict(node.metadata())}
                d = d[key]["contents"]
        return safe_json_dump(to_serialize)

    serialization_registry.register(
        StructureFamily.container, "application/json", serialize_json
    )



import awkward

from ..serialization.awkward import from_zipped_buffers, to_zipped_buffers
from ..structures.awkward import project_form
from .base import BaseClient
from .utils import export_util, handle_error


class AwkwardClient(BaseClient):
    def __repr__(self):
        # TODO Include some summary of the structure. Probably
        # lift __repr__ code from awkward itself here.
        return f"<{type(self).__name__}>"

    def write(self, container):
        structure = self.structure()
        components = (structure.form, structure.length, container)
        handle_error(
            self.context.http_client.put(
                self.item["links"]["full"],
                content=bytes(to_zipped_buffers(components, {})),
                headers={"Content-Type": "application/zip"},
            )
        )

    def read(self, slice=...):
        structure = self.structure()
        form = awkward.forms.from_dict(structure.form)
        typetracer, report = awkward.typetracer.typetracer_with_report(
            form,
        )
        proxy_array = awkward.Array(typetracer)
        awkward.typetracer.touch_data(proxy_array[slice])
        form_keys_touched = set(report.data_touched)
        projected_form = project_form(form, form_keys_touched)
        # The order is not important, but sort so that the request is deterministic.
        form_keys = sorted(list(form_keys_touched))
        content = handle_error(
            self.context.http_client.post(
                self.item["links"]["buffers"],
                headers={"Accept": "application/zip"},
                json=form_keys,
            )
        ).read()
        container = from_zipped_buffers(content, projected_form, structure.length)
        projected_array = awkward.from_buffers(
            projected_form,
            structure.length,
            container,
            allow_noncanonical_form=True,
        )
        return projected_array[slice]

    def __getitem__(self, slice):
        return self.read(slice=slice)

    def export(self, filepath, *, format=None):
        """
        Download data in some format and write to a file.

        Parameters
        ----------
        file: str or buffer
            Filepath or writeable buffer.
        format : str, optional
            If format is None and `file` is a filepath, the format is inferred
            from the name, like 'table.csv' implies format="text/csv". The format
            may be given as a file extension ("csv") or a media type ("text/csv").

        Examples
        --------

        Export as Parquet.

        >>> a.export("awkward.parquet")

        Export as JSON.

        >>> a.export("awkward.json")

        """
        return export_util(
            filepath,
            format,
            self.context.http_client.get,
            self.item["links"]["full"],
            params={},
        )



import hashlib
import uuid as uuid_module
from datetime import datetime

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import selectinload
from sqlalchemy.sql import func

from .base import Base
from .orm import APIKey, Identity, PendingSession, Principal, Role, Session

# This is the alembic revision ID of the database revision
# required by this version of Tiled.
REQUIRED_REVISION = "769180ce732e"
# This is list of all valid revisions (from current to oldest).
ALL_REVISIONS = [
    "769180ce732e",
    "c7bd2573716d",
    "4a9dfaba4a98",
    "56809bcbfcb0",
    "722ff4e4fcc7",
    "481830dd6c11",
]


async def create_default_roles(db):
    db.add_all(
        [
            Role(
                name="user",
                description="Default Role for users.",
                scopes=[
                    "read:metadata",
                    "read:data",
                    "create",
                    "write:metadata",
                    "write:data",
                    "apikeys",
                ],
            ),
            Role(
                name="admin",
                description="Role with elevated privileges.",
                scopes=[
                    "read:metadata",
                    "read:data",
                    "create",
                    "write:metadata",
                    "write:data",
                    "admin:apikeys",
                    "read:principals",
                    "write:principals",
                    "metrics",
                ],
            ),
        ]
    )
    await db.commit()


async def initialize_database(engine):
    # The definitions in .orm alter Base.metadata.
    from . import orm  # noqa: F401

    async with engine.connect() as conn:
        # Create all tables.
        await conn.run_sync(Base.metadata.create_all)

        # Initialize Roles table.
        async with AsyncSession(engine) as db:
            await create_default_roles(db)


async def purge_expired(db, cls):
    """
    Remove expired entries.
    """
    now = datetime.utcnow()
    num_expired = 0
    statement = (
        select(cls)
        .filter(cls.expiration_time.is_not(None))
        .filter(cls.expiration_time < now)
    )
    result = await db.execute(statement)
    for obj in result.scalars():
        num_expired += 1
        await db.delete(obj)
    if num_expired:
        await db.commit()
    return num_expired


async def create_user(db, identity_provider, id):
    user_role = (await db.execute(select(Role).filter(Role.name == "user"))).scalar()
    assert user_role is not None, "User role is missing from Roles table"
    principal = Principal(type="user", roles=[user_role])
    db.add(principal)
    await db.commit()
    identity = Identity(
        provider=identity_provider,
        id=id,
        principal_id=principal.id,
    )
    db.add(identity)
    await db.commit()
    refreshed_principal = (
        await db.execute(
            select(Principal)
            .filter(Principal.id == principal.id)
            .options(selectinload(Principal.identities))
        )
    ).scalar()
    return refreshed_principal


async def create_service(db, role):
    role_ = (await db.execute(select(Role).filter(Role.name == role))).scalar()
    if role_ is None:
        raise ValueError(f"Role named {role!r} is not found")
    principal = Principal(type="service", roles=[role_])
    db.add(principal)
    await db.commit()
    return principal


async def lookup_valid_session(db, session_id):
    if isinstance(session_id, int):
        # Old versions of tiled used an integer sid.
        # Reject any of those old sessions and force reauthentication.
        return None

    session = (
        await db.execute(
            select(Session)
            .options(
                selectinload(Session.principal).selectinload(Principal.roles),
                selectinload(Session.principal).selectinload(Principal.identities),
            )
            .filter(Session.uuid == uuid_module.UUID(hex=session_id))
        )
    ).scalar()
    if session is None:
        return None
    if (
        session.expiration_time is not None
        and session.expiration_time < datetime.utcnow()
    ):
        await db.delete(session)
        await db.commit()
        return None
    return session


async def lookup_valid_pending_session_by_device_code(db, device_code):
    hashed_device_code = hashlib.sha256(device_code).digest()
    pending_session = (
        await db.execute(
            select(PendingSession)
            .filter(PendingSession.hashed_device_code == hashed_device_code)
            .options(
                selectinload(PendingSession.session)
                .selectinload(Session.principal)
                .selectinload(Principal.identities),
            )
        )
    ).scalar()
    if pending_session is None:
        return None
    if (
        pending_session.expiration_time is not None
        and pending_session.expiration_time < datetime.utcnow()
    ):
        await db.delete(pending_session)
        await db.commit()
        return None
    return pending_session


async def lookup_valid_pending_session_by_user_code(db, user_code):
    pending_session = (
        await db.execute(
            select(PendingSession).filter(PendingSession.user_code == user_code)
        )
    ).scalar()
    if pending_session is None:
        return None
    if (
        pending_session.expiration_time is not None
        and pending_session.expiration_time < datetime.utcnow()
    ):
        await db.delete(pending_session)
        await db.commit()
        return None
    return pending_session


async def make_admin_by_identity(db, identity_provider, id):
    identity = (
        await db.execute(
            select(Identity)
            .options(selectinload(Identity.principal).selectinload(Principal.roles))
            .filter(Identity.id == id)
            .filter(Identity.provider == identity_provider)
        )
    ).scalar()
    if identity is None:
        principal = await create_user(db, identity_provider, id)
    else:
        principal = identity.principal
    admin_role = (await db.execute(select(Role).filter(Role.name == "admin"))).scalar()
    assert admin_role is not None, "Admin role is missing from Roles table"
    principal.roles.append(admin_role)
    await db.commit()
    return principal


async def lookup_valid_api_key(db, secret):
    """
    Look up an API key. Ensure that it is valid.
    """

    now = datetime.utcnow()
    hashed_secret = hashlib.sha256(secret).digest()
    api_key = (
        await db.execute(
            select(APIKey)
            .options(
                selectinload(APIKey.principal).selectinload(Principal.roles),
                selectinload(APIKey.principal).selectinload(Principal.identities),
            )
            .filter(APIKey.first_eight == secret.hex()[:8])
            .filter(APIKey.hashed_secret == hashed_secret)
        )
    ).scalar()
    if api_key is None:
        # No match
        validated_api_key = None
    elif (api_key.expiration_time is not None) and (api_key.expiration_time < now):
        # Match is expired. Delete it.
        await db.delete(api_key)
        await db.commit()
        validated_api_key = None
    elif api_key.principal is None:
        # The Principal for the API key no longer exists. Delete it.
        await db.delete(api_key)
        await db.commit()
        validated_api_key = None
    else:
        validated_api_key = api_key
    return validated_api_key


async def latest_principal_activity(db, principal):
    """
    The most recent time this Principal has logged in with an Identity,
    refreshed a Session, or used an APIKey.

    Note that activity that is authenticated using an access token is not
    captured here. As usual with JWTs, those requests do not interact with
    this database, for performance reasons. Therefore, this may lag actual
    activity by as much as the max age of an access token (default: 15
    minutes).
    """
    latest_identity_activity = (
        await db.execute(
            select(func.max(Identity.latest_login)).filter(
                Identity.principal_id == principal.id
            )
        )
    ).scalar()
    latest_session_activity = (
        await db.execute(
            select(func.max(Session.time_last_refreshed)).filter(
                Session.principal_id == principal.id
            )
        )
    ).scalar()
    latest_api_key_activity = (
        await db.execute(
            select(func.max(APIKey.latest_activity)).filter(
                APIKey.principal_id == principal.id
            )
        )
    ).scalar()
    all_activity = [
        latest_identity_activity,
        latest_api_key_activity,
        latest_session_activity,
    ]
    if all([t is None for t in all_activity]):
        return None
    return max(t for t in all_activity if t is not None)



import re
from pathlib import Path
from typing import List, Optional

import typer

serve_app = typer.Typer()

SQLITE_CATALOG_FILENAME = "catalog.db"
DATA_SUBDIRECTORY = "data"


@serve_app.command("directory")
def serve_directory(
    directory: str = typer.Argument(..., help="A directory to serve"),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help=("Log details of directory traversal and file registration."),
    ),
    watch: bool = typer.Option(
        False,
        "--watch",
        "-w",
        help="Update catalog when files are added, removed, or changed.",
    ),
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with "
            "this option selected."
        ),
    ),
    api_key: str = typer.Option(
        None,
        "--api-key",
        help=(
            "Set the single-user API key. "
            "By default, a random key is generated at startup and printed."
        ),
    ),
    keep_ext: bool = typer.Option(
        False,
        "--keep-ext",
        help=(
            "Serve a file like 'measurements.csv' as its full filepath with extension, "
            "instead of the default which would serve it as 'measurements'. "
            "This is discouraged because it leaks details about the storage "
            "format to the client, such that changing the storage in the future "
            "may break user (client-side) code."
        ),
    ),
    ext: List[str] = typer.Option(
        None,
        "--ext",
        help=(
            "Support custom file extension, mapping it to a known mimetype. "
            "Spell like '.tif=image/tiff'. Include the leading '.' in the file "
            "extension."
        ),
    ),
    mimetype_detection_hook: str = typer.Option(
        None,
        "--mimetype-hook",
        help=(
            "ADVANCED: Custom mimetype detection Python function. "
            "Expected interface: detect_mimetype(filepath, mimetype) -> mimetype "
            "Specify here as 'package.module:function'"
        ),
    ),
    adapters: List[str] = typer.Option(
        None,
        "--adapter",
        help=(
            "ADVANCED: Custom Tiled Adapter for reading a given format"
            "Specify here as 'mimetype=package.module:function'"
        ),
    ),
    walkers: List[str] = typer.Option(
        None,
        "--walker",
        help=(
            "ADVANCED: Custom Tiled Walker for traversing directories and "
            "grouping files. This is used in conjunction with Adapters that operate "
            "on groups of files. "
            "Specify here as 'package.module:function'"
        ),
    ),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
):
    "Serve a Tree instance from a directory of files."
    import tempfile

    temp_directory = Path(tempfile.TemporaryDirectory().name)
    temp_directory.mkdir()
    typer.echo(
        f"Creating catalog database at {temp_directory / SQLITE_CATALOG_FILENAME}",
        err=True,
    )
    database = f"sqlite+aiosqlite:///{Path(temp_directory, SQLITE_CATALOG_FILENAME)}"

    # Because this is a tempfile we know this is a fresh database and we do not
    # need to check its current state.
    # We _will_ go ahead and stamp it with a revision because it is possible the
    # user will copy it into a permanent location.

    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import stamp_head
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import initialize_database

    engine = create_async_engine(database)
    asyncio.run(initialize_database(engine))
    stamp_head(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database)

    from ..catalog import from_uri
    from ..server.app import build_app, print_admin_api_key_if_generated

    server_settings = {}
    if keep_ext:
        from ..adapters.files import identity

        key_from_filename = identity
    else:
        key_from_filename = None
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes

    from logging import StreamHandler

    from ..catalog.register import logger as register_logger
    from ..catalog.register import register
    from ..catalog.register import watch as watch_

    mimetypes_by_file_ext = {}
    EXT_PATTERN = re.compile(r"(.*) *= *(.*)")
    for item in ext or []:
        match = EXT_PATTERN.match(item)
        if match is None:
            raise ValueError(
                f"Failed parsing --ext option {item}, expected format '.ext=mimetype'"
            )
        ext, mimetype = match.groups()
        mimetypes_by_file_ext[ext] = mimetype
    adapters_by_mimetype = {}
    ADAPTER_PATTERN = re.compile(r"(.*) *= *(.*)")
    for item in adapters or []:
        match = ADAPTER_PATTERN.match(item)
        if match is None:
            raise ValueError(
                f"Failed parsing --adapter option {item}, expected format 'mimetype=package.module:obj'"
            )
        mimetype, obj_ref = match.groups()
        adapters_by_mimetype[mimetype] = obj_ref
    catalog_adapter = from_uri(
        database,
        readable_storage=[directory],
        adapters_by_mimetype=adapters_by_mimetype,
    )
    typer.echo(f"Indexing '{directory}' ...")
    if verbose:
        register_logger.addHandler(StreamHandler())
        register_logger.setLevel("INFO")
    web_app = build_app(
        catalog_adapter,
        {
            "allow_anonymous_access": public,
            "single_user_api_key": api_key,
        },
        server_settings,
    )
    if watch:

        async def walk_and_serve():
            import anyio

            event = anyio.Event()
            asyncio.create_task(
                watch_(
                    catalog_adapter,
                    directory,
                    initial_walk_complete_event=event,
                    mimetype_detection_hook=mimetype_detection_hook,
                    mimetypes_by_file_ext=mimetypes_by_file_ext,
                    adapters_by_mimetype=adapters_by_mimetype,
                    walkers=walkers,
                    key_from_filename=key_from_filename,
                )
            )
            await event.wait()
            typer.echo("Initial indexing complete. Starting server...")
            print_admin_api_key_if_generated(web_app, host=host, port=port)

            import uvicorn

            config = uvicorn.Config(web_app, host=host, port=port)
            server = uvicorn.Server(config)
            await server.serve()

        asyncio.run(walk_and_serve())
    else:
        asyncio.run(
            register(
                catalog_adapter,
                directory,
                mimetype_detection_hook=mimetype_detection_hook,
                mimetypes_by_file_ext=mimetypes_by_file_ext,
                adapters_by_mimetype=adapters_by_mimetype,
                walkers=walkers,
                key_from_filename=key_from_filename,
            )
        )

        typer.echo("Indexing complete. Starting server...")
        print_admin_api_key_if_generated(web_app, host=host, port=port)

        import uvicorn

        uvicorn.run(web_app, host=host, port=port)


def serve_catalog(
    database: str = typer.Argument(
        None, help="A filepath or database URI, e.g. 'catalog.db'"
    ),
    read: List[str] = typer.Option(
        None,
        "--read",
        "-r",
        help="Locations that the server may read from",
    ),
    write: str = typer.Option(
        None,
        "--write",
        "-w",
        help="Location that the server may write to",
    ),
    temp: bool = typer.Option(
        False,
        "--temp",
        help="Make a new catalog in a temporary directory.",
    ),
    init: bool = typer.Option(
        False,
        "--init",
        help="Initialize a new catalog database.",
    ),
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with "
            "this option selected."
        ),
    ),
    api_key: str = typer.Option(
        None,
        "--api-key",
        help=(
            "Set the single-user API key. "
            "By default, a random key is generated at startup and printed."
        ),
    ),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
    scalable: bool = typer.Option(
        False,
        "--scalable",
        help=(
            "This verifies that the configuration is compatible with scaled (multi-process) deployments."
        ),
    ),
):
    "Serve a catalog."
    import urllib.parse

    from ..catalog import from_uri
    from ..server.app import build_app, print_admin_api_key_if_generated

    parsed_database = urllib.parse.urlparse(database)
    if parsed_database.scheme in ("", "file"):
        database = f"sqlite+aiosqlite:///{parsed_database.path}"

    if temp:
        if database is not None:
            typer.echo(
                "The option --temp was set but a database was also provided. "
                "Do one or the other.",
                err=True,
            )
            raise typer.Abort()
        import tempfile

        directory = Path(tempfile.TemporaryDirectory().name)
        directory.mkdir()
        typer.echo(
            f"Creating catalog database at {directory / SQLITE_CATALOG_FILENAME}",
            err=True,
        )
        typer.echo(
            f"Creating writable catalog data directory at {directory / DATA_SUBDIRECTORY}",
            err=True,
        )
        database = f"sqlite+aiosqlite:///{Path(directory, SQLITE_CATALOG_FILENAME)}"

        # Because this is a tempfile we know this is a fresh database and we do not
        # need to check its current state.
        # We _will_ go ahead and stamp it with a revision because it is possible the
        # user will copy it into a permanent location.

        import asyncio

        from sqlalchemy.ext.asyncio import create_async_engine

        from ..alembic_utils import stamp_head
        from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
        from ..catalog.core import initialize_database

        engine = create_async_engine(database)
        asyncio.run(initialize_database(engine))
        stamp_head(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database)

        if write is None:
            write = directory / DATA_SUBDIRECTORY
            write.mkdir()
        # TODO Hook into server lifecycle hooks to delete this at shutdown.
    elif database is None:
        typer.echo(
            """A catalog must be specified. Either use a temporary catalog:

    tiled serve catalog --temp

or initialize a new catalog, e.g.

    tiled catalog init catalog.db
    tiled serve catalog catalog.db

or use an existing one:

    tiled serve catalog catalog.db
""",
            err=True,
        )
        raise typer.Abort()

    if write is None:
        typer.echo(
            "This catalog will be served as read-only. "
            "To make it writable, specify a writable directory with --write.",
            err=True,
        )
    server_settings = {}
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes
    tree = from_uri(
        database,
        writable_storage=write,
        readable_storage=read,
        init_if_not_exists=init,
    )
    web_app = build_app(
        tree,
        {
            "allow_anonymous_access": public,
            "single_user_api_key": api_key,
        },
        server_settings,
        scalable=scalable,
    )
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


serve_app.command("catalog")(serve_catalog)


@serve_app.command("pyobject")
def serve_pyobject(
    object_path: str = typer.Argument(
        ..., help="Object path, as in 'package.subpackage.module:object_name'"
    ),
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with this "
            "option selected."
        ),
    ),
    api_key: str = typer.Option(
        None,
        "--api-key",
        help=(
            "Set the single-user API key. "
            "By default, a random key is generated at startup and printed."
        ),
    ),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
    scalable: bool = typer.Option(
        False,
        "--scalable",
        help=(
            "This verifies that the configuration is compatible with scaled (multi-process) deployments."
        ),
    ),
):
    "Serve a Tree instance from a Python module."
    from ..server.app import build_app, print_admin_api_key_if_generated
    from ..utils import import_object

    tree = import_object(object_path)
    server_settings = {}
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes
    web_app = build_app(
        tree,
        {
            "allow_anonymous_access": public,
            "single_user_api_key": api_key,
        },
        server_settings,
        scalable=scalable,
    )
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


@serve_app.command("demo")
def serve_demo(
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
):
    "Start a public server with example data."
    from ..server.app import build_app, print_admin_api_key_if_generated
    from ..utils import import_object

    EXAMPLE = "tiled.examples.generated:tree"
    tree = import_object(EXAMPLE)
    web_app = build_app(tree, {"allow_anonymous_access": True}, {})
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


@serve_app.command("config")
def serve_config(
    config_path: Path = typer.Argument(
        None,
        help=(
            "Path to a config file or directory of config files. "
            "If None, check environment variable TILED_CONFIG. "
            "If that is unset, try default location ./config.yml."
        ),
    ),
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with this "
            "option selected."
        ),
    ),
    api_key: str = typer.Option(
        None,
        "--api-key",
        help=(
            "Set the single-user API key. "
            "By default, a random key is generated at startup and printed."
        ),
    ),
    host: str = typer.Option(
        None,
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`. Uses value in config by default."
        ),
    ),
    port: int = typer.Option(
        None, help="Bind to a socket with this port. Uses value in config by default."
    ),
    scalable: bool = typer.Option(
        False,
        "--scalable",
        help=(
            "This verifies that the configuration is compatible with scaled (multi-process) deployments."
        ),
    ),
):
    "Serve a Tree as specified in configuration file(s)."
    import os

    from ..config import parse_configs

    config_path = config_path or os.getenv("TILED_CONFIG", "config.yml")
    try:
        parsed_config = parse_configs(config_path)
    except Exception as err:
        typer.echo(str(err), err=True)
        raise typer.Abort()

    # Let --public flag override config.
    if public:
        if "authentication" not in parsed_config:
            parsed_config["authentication"] = {}
        parsed_config["authentication"]["allow_anonymous_access"] = True
    # Let --api-key flag override config.
    if api_key:
        if "authentication" not in parsed_config:
            parsed_config["authentication"] = {}
        parsed_config["authentication"]["single_user_api_key"] = api_key

    # Delay this import so that we can fail faster if config-parsing fails above.

    from ..server.app import (
        build_app_from_config,
        logger,
        print_admin_api_key_if_generated,
    )

    # Extract config for uvicorn.
    uvicorn_kwargs = parsed_config.pop("uvicorn", {})
    # If --host is given, it overrides host in config. Same for --port.
    uvicorn_kwargs["host"] = host or uvicorn_kwargs.get("host", "127.0.0.1")
    uvicorn_kwargs["port"] = port or uvicorn_kwargs.get("port", 8000)

    # This config was already validated when it was parsed. Do not re-validate.
    logger.info(f"Using configuration from {Path(config_path).absolute()}")

    if root_path := uvicorn_kwargs.get("root_path", ""):
        parsed_config["root_path"] = root_path

    web_app = build_app_from_config(
        parsed_config, source_filepath=config_path, scalable=scalable
    )
    print_admin_api_key_if_generated(
        web_app, host=uvicorn_kwargs["host"], port=uvicorn_kwargs["port"]
    )

    # Likewise, delay this import.

    import uvicorn

    uvicorn.run(web_app, **uvicorn_kwargs)



"""Add 'write:principals' scope to admin

Revision ID: 769180ce732e
Revises: c7bd2573716d
Create Date: 2023-12-12 17:57:56.388145

"""
from alembic import op
from sqlalchemy.orm.session import Session

from tiled.authn_database.orm import Role

# revision identifiers, used by Alembic.
revision = "769180ce732e"
down_revision = "c7bd2573716d"
branch_labels = None
depends_on = None


SCOPE = "write:principals"


def upgrade():
    """
    Add 'write:principals' scope to default 'admin' Role.
    """
    connection = op.get_bind()
    with Session(bind=connection) as db:
        role = db.query(Role).filter(Role.name == "admin").first()
        scopes = role.scopes.copy()
        scopes.append(SCOPE)
        role.scopes = scopes
        db.commit()


def downgrade():
    """
    Remove new scopes from Roles, if present.
    """
    connection = op.get_bind()
    with Session(bind=connection) as db:
        role = db.query(Role).filter(Role.name == "admin").first()
        scopes = role.scopes.copy()
        if SCOPE in scopes:
            scopes.remove(SCOPE)
        role.scopes = scopes
        db.commit()



import contextlib
import os.path
import tempfile

from alembic import command
from alembic.config import Config
from alembic.runtime import migration


def write_alembic_ini(alembic_ini_template_path, alembic_dir, path, database_uri):
    """Write a complete alembic.ini from our template.

    Parameters
    ----------
    alembic_ini_template_path: str
    alembic_dir: str
    path : str
        path to the alembic.ini file that should be written.
    db_url : str
        The SQLAlchemy database url, e.g. `sqlite+aiosqlite:///tiled.sqlite`.
    """
    with open(alembic_ini_template_path) as f:
        alembic_ini_tpl = f.read()

    with open(path, "w") as f:
        f.write(
            alembic_ini_tpl.format(
                migration_script_directory=alembic_dir,
                # If there are any %s in the URL, they should be replaced with %%, since ConfigParser
                # by default uses %() for substitution. You'll get %s in your URL when you have usernames
                # with special chars (such as '@') that need to be URL encoded. URL Encoding is done with %s.
                # YAY for nested templates?
                database_uri=str(database_uri).replace("%", "%%"),
            )
        )


@contextlib.contextmanager
def temp_alembic_ini(alembic_ini_template_path, alembic_dir, database_uri):
    """
    Context manager for temporary alembic configuration file

    Temporarily write an alembic.ini file for use with alembic migration scripts.
    Context manager yields alembic.ini path.

    Parameters
    ----------
    alembic_ini_template_path: str
    alembic_dir: str
    datbase_uri : str
        The SQLAlchemy database url, e.g. `sqlite+aiosqlite:///tiled.sqlite`.

    Returns
    -------
    alembic_ini: str
        The path to the temporary alembic.ini that we have created.
        This file will be cleaned up on exit from the context manager.
    """
    with tempfile.TemporaryDirectory() as td:
        alembic_ini = os.path.join(td, "alembic.ini")
        write_alembic_ini(
            alembic_ini_template_path, alembic_dir, alembic_ini, database_uri
        )
        yield alembic_ini


def stamp_head(alembic_ini_template_path, alembic_dir, engine_url):
    """
    Upgrade schema to the specified revision.
    """
    with temp_alembic_ini(
        alembic_ini_template_path, alembic_dir, engine_url
    ) as alembic_ini:
        alembic_cfg = Config(alembic_ini)
        command.stamp(alembic_cfg, "head")


def upgrade(alembic_ini_template_path, alembic_dir, engine_url, revision):
    """
    Upgrade schema to the specified revision.
    """
    with temp_alembic_ini(
        alembic_ini_template_path, alembic_dir, engine_url
    ) as alembic_ini:
        alembic_cfg = Config(alembic_ini)
        command.upgrade(alembic_cfg, revision)


def downgrade(alembic_ini_template_path, alembic_dir, engine_url, revision):
    """
    Downgrade schema to the specified revision.
    """
    with temp_alembic_ini(
        alembic_ini_template_path, alembic_dir, engine_url
    ) as alembic_ini:
        alembic_cfg = Config(alembic_ini)
        command.downgrade(alembic_cfg, revision)


class UnrecognizedDatabase(Exception):
    pass


class UninitializedDatabase(Exception):
    pass


class DatabaseUpgradeNeeded(Exception):
    pass


async def get_current_revision(engine, known_revisions):
    redacted_url = engine.url._replace(password="[redacted]")
    async with engine.connect() as conn:
        context = await conn.run_sync(migration.MigrationContext.configure)
        heads = await conn.run_sync(lambda conn: context.get_current_heads())
    if heads == ():
        return None
    elif len(heads) != 1:
        raise UnrecognizedDatabase(
            f"This database {redacted_url} is stamped with an alembic revisions {heads}. "
            "It looks like Tiled has been configured to connect to a database "
            "already populated by some other application (not Tiled) or else "
            "its database is in a corrupted state."
        )
    (revision,) = heads
    if revision not in known_revisions:
        raise UnrecognizedDatabase(
            f"The datbase {redacted_url} has an unrecognized revision {revision}. "
            "It may have been created by a newer version of Tiled."
        )
    return revision


async def check_database(engine, required_revision, known_revisions):
    revision = await get_current_revision(engine, known_revisions)
    redacted_url = engine.url._replace(password="[redacted]")
    if revision is None:
        raise UninitializedDatabase(
            f"The database {redacted_url} has no revision stamp. It may be empty. "
            "It can be initialized with `initialize_database(engine)`."
        )
    elif revision != required_revision:
        raise DatabaseUpgradeNeeded(
            f"The database {redacted_url} has revision {revision} and "
            f"needs to be upgraded to revision {required_revision}."
        )



from ..profiles import (
    create_profile,
    delete_profile,
    get_default_profile_name,
    list_profiles,
    load_profiles,
    set_default_profile_name,
)


def test_integration(tmp_profiles_dir):
    URI = "http://example.com/api/v1"
    create_profile(name="test", uri=URI)
    assert "test" in list_profiles()
    _, profile_content = load_profiles()["test"]
    assert profile_content["uri"] == URI
    assert get_default_profile_name() is None
    set_default_profile_name("test")
    assert get_default_profile_name() == "test"
    set_default_profile_name(None)
    assert get_default_profile_name() is None
    delete_profile("test")
    assert "test" not in list_profiles()



import builtins
import hashlib

import tifffile

from ..server.object_cache import with_object_cache
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.core import StructureFamily
from ..utils import path_from_uri


class TiffAdapter:
    """
    Read a TIFF file.

    Examples
    --------

    >>> TiffAdapter("path/to/file.tiff")
    """

    structure_family = StructureFamily.array

    def __init__(
        self,
        data_uri,
        *,
        structure=None,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        if not isinstance(data_uri, str):
            raise Exception
        filepath = path_from_uri(data_uri)
        self._file = tifffile.TiffFile(filepath)
        self._cache_key = (type(self).__module__, type(self).__qualname__, filepath)
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        self.access_policy = access_policy
        if structure is None:
            if self._file.is_shaped:
                shape = tuple(self._file.shaped_metadata[0]["shape"])
            else:
                arr = with_object_cache(self._cache_key, self._file.asarray)
                shape = arr.shape
            structure = ArrayStructure(
                shape=shape,
                chunks=tuple((dim,) for dim in shape),
                data_type=BuiltinDtype.from_numpy_dtype(self._file.series[0].dtype),
            )
        self._structure = structure

    def metadata(self):
        # This contains some enums, but Python's built-in JSON serializer
        # handles them fine (converting  to str or int as appropriate).
        d = {tag.name: tag.value for tag in self._file.pages[0].tags.values()}
        d.update(self._provided_metadata)
        return d

    def read(self, slice=None):
        # TODO Is there support for reading less than the whole array
        # if we only want a slice? I do not think that is possible with a
        # single-page TIFF but I'm not sure. Certainly it *is* possible for
        # multi-page TIFFs.
        arr = with_object_cache(self._cache_key, self._file.asarray)
        if slice is not None:
            arr = arr[slice]
        return arr

    def read_block(self, block, slice=None):
        # For simplicity, this adapter always treat a single TIFF file as one
        # chunk. This could be relaxed in the future.
        if sum(block) != 0:
            raise IndexError(block)

        arr = with_object_cache(self._cache_key, self._file.asarray)
        if slice is not None:
            arr = arr[slice]
        return arr

    def structure(self):
        return self._structure


class TiffSequenceAdapter:
    structure_family = "array"

    @classmethod
    def from_uris(
        cls,
        data_uris,
        structure=None,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        filepaths = [path_from_uri(data_uri) for data_uri in data_uris]
        seq = tifffile.TiffSequence(filepaths)
        return cls(
            seq,
            structure=structure,
            specs=specs,
            metadata=metadata,
            access_policy=access_policy,
        )

    def __init__(
        self,
        seq,
        *,
        structure=None,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        self._seq = seq
        self._cache_key = (
            type(self).__module__,
            type(self).__qualname__,
            hashlib.md5(str(self._seq.files).encode()).hexdigest(),
        )
        # TODO Check shape, chunks against reality.
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        self.access_policy = access_policy
        if structure is None:
            shape = (len(self._seq), *self.read(slice=0).shape)
            structure = ArrayStructure(
                shape=shape,
                # one chunks per underlying TIFF file
                chunks=((1,) * shape[0], (shape[1],), (shape[2],)),
                # Assume all files have the same data type
                data_type=BuiltinDtype.from_numpy_dtype(self.read(slice=0).dtype),
            )
        self._structure = structure

    def metadata(self):
        # TODO How to deal with the many headers?
        return self._provided_metadata

    def read(self, slice=None):
        """Return a numpy array

        Receives a sequence of values to select from a collection of tiff files that were saved in a folder
        The input order is defined as files --> X slice --> Y slice
        read() can receive one value or one slice to select all the data from one file or a sequence of files;
        or it can receive a tuple of up to three values (int or slice) to select a more specific sequence of pixels
        of a group of images
        """

        if slice is None:
            return with_object_cache(self._cache_key, self._seq.asarray)
        if isinstance(slice, int):
            # e.g. read(slice=0)
            return with_object_cache(
                self._cache_key + (slice,),
                tifffile.TiffFile(self._seq.files[slice]).asarray,
            )
        # e.g. read(slice=(...))
        if isinstance(slice, tuple):
            if len(slice) == 0:
                return with_object_cache(self._cache_key, self._seq.asarray)
            image_axis, *the_rest = slice
            # Could be int or slice
            # (0, slice(...)) or (0,....) are converted to a list
            if isinstance(image_axis, int):
                # e.g. read(slice=(0, ....))
                arr = with_object_cache(
                    self._cache_key + (image_axis,),
                    tifffile.TiffFile(self._seq.files[image_axis]).asarray,
                )
            if isinstance(image_axis, builtins.slice):
                if image_axis.start is None:
                    slice_start = 0
                else:
                    slice_start = image_axis.start
                if image_axis.step is None:
                    slice_step = 1
                else:
                    slice_step = image_axis.step

                arr = with_object_cache(
                    self._cache_key + (slice_start, image_axis.stop, slice_step),
                    tifffile.TiffSequence(
                        self._seq.files[
                            slice_start : image_axis.stop : slice_step  # noqa: E203
                        ]
                    ).asarray,
                )
            arr = arr[tuple(the_rest)]
            return arr
        if isinstance(slice, builtins.slice):
            # Check for start and step which can be optional
            if slice.start is None:
                slice_start = 0
            else:
                slice_start = slice.start
            if slice.step is None:
                slice_step = 1
            else:
                slice_step = slice.step

            arr = with_object_cache(
                self._cache_key + (slice_start, slice.stop, slice_step),
                tifffile.TiffSequence(
                    self._seq.files[slice_start : slice.stop : slice_step]  # noqa: E203
                ).asarray,
            )
            return arr

    def read_block(self, block, slice=None):
        if block[1:] != (0, 0):
            raise IndexError(block)
        arr = self.read(builtins.slice(block[0], block[0] + 1))
        if slice is not None:
            arr = arr[slice]
        return arr

    def structure(self):
        return self._structure
